/*cdSPEC returns the spectrum needed for Keith Arnaud's XSPEC */
#include "cddefines.h"
#include "cddrive.h"
#include "physconst.h"
#include "sphere.h"
#include "radius.h"
#include "rfield.h"

/* 
 * this routine returns the spectrum needed for Keith Arnaud's XSPEC
 * X-Ray analysis code.  It should be called after cdDrive has successfully
 * computed a model.  the calling routine must ensure that the three vectors
 * have enough space to store the resulting spectrum, 
 * given the bounds and energy resolution 
 */

/* pointer to array energy in Cloudy grids -  this has file scope */
static long int ip;

/* energies of lower and upper bound of XSPEC cell we are considering */
static double Elo , Ehi;
/* interpolate on cloudy's predicted spectrum to get results on XSPEC grid */
double Spec_cont( float cont[] );

void cdSPEC( 
	/* option - the type of spectrum to be returned
	 * 1	the incident continuum 4\pi nuJ_nu, , erg cm-2 s-1
	 *
	 * 2	the attenuated incident continuum, same units
	 * 3	the reflected continuum, same units
	 *
	 * 4	diffuse continuous emission outward direction
	 * 5	diffuse continuous emission, reflected
	 *
	 * 6	collisional+recombination lines, outward
	 * 7	collisional+recombination lines, reflected
	 * 
	 * 8	pumped lines, outward
	 * 9	pumped lines, reflected
	 *
	 *		all lines and continuum emitted by the cloud assume full coverage of 
	 *		continuum source */
	int nOption ,

	/* the energy of the lower edge of each cell 
	 * (in Ryd to be consistent with all the rest of Cloudy) */
	double EnergyLow[] , 

	/* the number of cells + 1*/
	long int nEnergy ,

	/* the returned spectrum, same size is two energy spectra (see option), returns nEnergy -1 pts */
	double ReturnedSpectrum[] )

{
	/* this will be used as pointer to continuum array */
	float *cont , refac;
	long int ncell , j;
	/* flag saying whether we will need to free cont at the end */
	int lgFREE;

#	ifdef DEBUG_FUN
	fputs( "<+>cdSPEC()\n", debug_fp );
#	endif

	if( nOption == 1 )
	{
		/* this is the incident continuum, col 2 of punch continuum command */
		cont = rfield.FluxSave;
		lgFREE = FALSE;
	}
	else if( nOption == 2 )
	{
		/* the attenuated transmitted continuum, no diffuse emission,
		 * col 3 of punch continuum command */
		cont = rfield.flux;
		lgFREE = FALSE;
	}
	else if( nOption == 3 )
	{
		/* reflected incident continuum, col 6 of punch continuum command */
		lgFREE = FALSE;
		cont = rfield.ConRefIncid ;
	}
	else if( nOption == 4 )
	{
		/* diffuse continuous emission outward direction  */
		if( (cont = malloc( sizeof(float)*NCELL ) ) == NULL)
		{
			fprintf(ioQQQ," cdSPEC could not malloc nOption 4 \n");
			cdEXIT(1);
		}
		/* need to free the vector once done */
		lgFREE = TRUE;
		refac = (float)radius.r1r0sq*sphere.covgeo;
		for( j=0; j<rfield.nflux ; ++j)
		{
			cont[j] = (rfield.ConOutNoInter[j] + rfield.ConOutInter[j] )*refac ;
		}
	}
	else if( nOption == 5 )
	{
		/* reflected diffuse continuous emission */
		if( (cont = malloc( sizeof(float)*NCELL ) ) == NULL)
		{
			fprintf(ioQQQ," cdSPEC could not malloc nOption 5 \n");
			cdEXIT(1);
		}
		/* need to free the vector once done */
		lgFREE = TRUE;
		refac = (float)radius.r1r0sq*sphere.covgeo;
		for( j=0; j<rfield.nflux ; ++j)
		{
			cont[j] = rfield.ConRefDiff[j]*refac  ;
		}
	}
	else if( nOption == 6 )
	{
		/* all outward lines,   */
		if( (cont = malloc( sizeof(float)*NCELL ) ) == NULL)
		{
			fprintf(ioQQQ," cdSPEC could not malloc nOption 6 \n");
			cdEXIT(1);
		}
		/* need to free the vector once done */
		lgFREE = TRUE;
		/* correct back to inner radius */
		refac = (float)radius.r1r0sq*sphere.covgeo;
		for( j=0; j<rfield.nflux ; ++j)
		{
			cont[j] = rfield.outlin[j] *rfield.widflx[j]/rfield.anu[j]*refac ;
		}
	}
	else if( nOption == 7 )
	{
		/* all reflected lines */
		if( sphere.lgSphere )
		{
			refac = 0.;
		}
		else
		{
			refac = 1.;
		}

		if( (cont = malloc( sizeof(float)*NCELL ) ) == NULL)
		{
			fprintf(ioQQQ," cdSPEC could not malloc nOption 7 \n");
			cdEXIT(1);
		}
		/* need to free the vector once done */
		lgFREE = TRUE;
		for( j=0; j<rfield.nflux ; ++j)
		{
			cont[j] = rfield.reflin[j] *rfield.widflx[j]/rfield.anu[j]*refac ;
		}
	}
	else
	{
		fprintf(ioQQQ," cdSPEC called with impossible nOption (%i)\n", nOption);
		puts( "[Stop in cdSPEC]" );
		cdEXIT(1);exit(1);
	}

	/* this is array index used in Spec_cont */
	ip = 0;
	/* now generate the continua */
	for( ncell = 0; ncell < nEnergy-1; ++ncell )
	{
		/* Spec_cont will find the spectrum between these two energies */
		Elo = EnergyLow[ncell];
		Ehi = EnergyLow[ncell+1];
		ReturnedSpectrum[ncell] = Spec_cont( cont );
	}

	/* need to free the vector once done if this flag was set */
	if( lgFREE )
	{
		free(cont) ;
	}
#	ifdef DEBUG_FUN
	fputs( " <->cdSPEC()\n", debug_fp );
#	endif
	return;
}

/* interpolate on cloudy's predicted spectrum to get results on XSPEC grid */
double Spec_cont( float cont[] )
{
	double sum,
		EcHi ,
		EcLo ;

	/* return zero if continuum does not go this high */
	if( ip >= rfield.nflux )
		return 0.;

	/* now skip out to where continuum starts, often we will already be there
	 * if this is a successive call */
	while( (Elo >= (rfield.anu[ip]+rfield.widflx[ip]/2.)  ) && ip < rfield.nflux )
		++ip;

	if( rfield.anu[ip]+rfield.widflx[ip]/2. > Ehi )
	{
		/* this is case where requested cell is within a single Cloudy cell */
		return( cont[ip] /rfield.widflx[ip] *rfield.anu2[ip]*EN1RYD );
	}
	else
	{
		/* this is case where we will add on more than one Cloudy cell */
		sum = cont[ip] * rfield.anu2[ip]* EN1RYD ;
		/* energy of upper edge of Cloudy cell */
		EcHi = ( rfield.anu[ip] + rfield.widflx[ip]/2. + rfield.anu[ip+1] - rfield.widflx[ip+1]/2.)/2.;
		/* mutiply this by energy width of this first cell that counts to final integral */
		sum *= (EcHi - Elo )/rfield.widflx[ip];
	}

	/* increment index since we will now work on next cell */
	++ip;

	/* now add up total cells, this loop for cloudy cells totally within desired bins,
	 * we will divide by energy width later, so factor of widflx is missing here */
	while( Ehi > rfield.anu[ip]+rfield.widflx[ip]/2. )
	{
		sum += cont[ip] * rfield.anu2[ip]*EN1RYD ;
		++ip;
	}

	/* energy of upper edge of Cloudy cell */
	EcLo = ( rfield.anu[ip] - rfield.widflx[ip]/2. + rfield.anu[ip-1] + rfield.widflx[ip-1]/2.)/2.;

	/* now finish up with last cell, whose upper bound is higher than desired bin */
	sum += cont[ip] * rfield.anu2[ip] * EN1RYD *
		(Ehi - EcLo )/rfield.widflx[ip];

	/* now divide by total energy width */
	sum /= (Ehi - Elo );

	return(sum);
}
