/*grain main routine to converge grains thermal solution */
#include <string.h>
#include <limits.h>
#include <float.h>
#include "cddefines.h"
#include "physconst.h"
#include "atomcwgt.h"
#include "heating.h"
#include "rfield.h"
#include "hmi.h"
#include "trace.h"
#include "converge.h"
#include "ipoint.h"
#include "phycon.h"
#include "taulines.h"
#include "heavy.h"
#include "h.h"
#include "negdrg.h"
#include "interpolate.h"
#include "hevmolec.h"
#include "ionfracs.h"
#include "itercnt.h"
#include "grnvrydpth.h"
#include "grainvar.h"
#include "grain.h"
#include "abundances.h"

#ifdef INCLUDE_OLD_GRAINS
#include "mie.h"
#include "dustop.h"
#endif

#define NINT(X) ((long)((X) < 0. ? (X)-0.5 : (X)+0.5))

/*================================================================================*/
/* these are used for setting up grain emissivities in InitEmissivities() */
#define GRAIN_TMIN 1.e-3
#define GRAIN_TMID 5.e3
#define GRAIN_TMAX 1.2e9

/* NTOP is number of bins for temps between GRAIN_TMID and GRAIN_TMAX */
#define NTOP 20L

/*================================================================================*/
/* miscellaneous grain physics */

/* constant for thermionic emissions, 7.501e20 e/cm^2/s/K^2 */
#define	THERMCONST (PI4*ELECTRON_MASS*POW2(BOLTZMANN)/POW3(HPLANCK))

/* sticking probabilities */
#define	STICK_ELEC	0.5
#define	STICK_ION	1.0

/* mean free path for electron penetrating grain, in cm */
#define MEAN_PATH	1.e-7

/*================================================================================*/
/* these are used when iterating the grain charge in GrainCharge() */
#define	DMINX	      (10.f*FLT_EPSILON)
#define	TOLER          0.005
#define BIG_CHANGE     5.
#define BIG_POTENTIAL  5.f
#define GOOD_PROGRESS  2.

#define BRACKET_MAX    50

/* finite range for bandgap to kick in, in Ryd */
#define DELTA_BANDGAP  1.e-4f

/* assumed 1-sigma width of grain charge distribution, in elem. units, used in GrainScreen */
#define SIG_DUSTZ      2.

/*================================================================================*/
/* definitions relating to quantum heating routines */

/* this is the minimum number of bins for quantum heating calculation to be valid */
#define NQMIN 10

/* this is the lowest value for dPdlnT that should be included in the modeling */
#define PROB_CUTOFF_LO 1.e-15
#define PROB_CUTOFF_HI 1.e-15
#define SAFETY         1.e+8

/* during the first NSTARTUP steps, use special algorithm to calculate stepsize */
#define NSTARTUP 5

/* if the average number of multiple events is above this number
 * don't even try quantum heating. */
#define MAX_EVENTS 100.

/* maximum number of tries for quantum heating routine */
#define LOOP_MAX 10

/* if all else fails, divide temp estimate by this factor */
#define DEF_FAC 5.

/* total probability for all bins should not deviate more than this from 1. */
#define PROB_TOL 0.05

/*================================================================================*/
/* these data define the enthalpy function for silicates
 * derived from: Guhathakurta & Draine, 1989, ApJ 345, 230 *
 * coefficients converted to rydberg energy units, per unit atom
 * assuming a density of 3.3 g/cm^3 and pure MgSiFeO4.
 * this is not right, but the result is correct because number
 * of atoms will be calculated using the same assumption. */

/* this is the specific density of silicate in g/cm^3 */
#define DEN_SIL     3.30

/* these are the mean molecular weights per atom for MgSiFeO4 and SiO2 in amu */
#define MW_SIL      24.6051
#define MW_SIO2     20.0283

static double tlim[5]={0.,50.,150.,500.,DBL_MAX};
static double power[4]={3.00,2.30,1.68,1.00};
static double cval[4]={
	1.40e3/3.00/DEN_SIL*ATOMIC_MASS_UNIT*MW_SIL/EN1RYD,
	2.20e4/2.30/DEN_SIL*ATOMIC_MASS_UNIT*MW_SIL/EN1RYD,
	4.80e5/1.68/DEN_SIL*ATOMIC_MASS_UNIT*MW_SIL/EN1RYD,
	3.41e7/1.00/DEN_SIL*ATOMIC_MASS_UNIT*MW_SIL/EN1RYD};


/* initialize grain emissivities */
static void InitEmissivities(void);
/* PlanckIntegral compute total radiative cooling due to large grains */
static double PlanckIntegral(double,long,long);
/* iterate grain charge and temperature */
static void GrainChargeTemp(void);
/* GrainCharge compute grains charge */
static void GrainCharge(long,double*);
/* bracket the solution for the grain potential */
static void BracketGrainPot(long,/*@in@*/float*,/*@out@*/double*,/*@in@*/float*,
			    /*@out@*/double*,/*@out@*/double*);
/* GrainElecRecomb compute electron recombination onto grains surface */
static double GrainElecRecomb(long int nd);
/* GrainElecEmis computes electron loss rate for the grain species
 * due to photoelectric effect, positive ion recombinations on the
 * grain surface and thermionic emissions */
static double GrainElecEmis(long int nd,/*@out@*/double*);
/* correction factors for grain charge screening (including image potential
 * to correct for polarization of the grain as charged particle approaches). */
static void GrainScreen(long,long,double*,double*);
/* update items that depend on grain potential */
static void UpdatePot(long);
/* GrainTemperature computes grains temperature, and gas cooling */
static void GrainTemperature(long,/*@out@*/float*,/*@out@*/double*,/*@out@*/double*,
			     /*@out@*/double*,/*@out@*/double*,/*@out@*/double*);
/* GrainCollHeating computes grains collisional heating cooling */
static void GrainCollHeating(long,/*@out@*/float*,/*@out@*/float*);
/* GrainDrift computes grains drift velocity */
static void GrainDrift(void);
/* worker routine, this implements the algorithm of Guhathakurtha & Draine */
static void qheat1(/*@out@*/double[],/*@out@*/double[],/*@out@*/double[],/*@out@*/long*,long,
		   /*@out@*/double*,/*@out@*/int*,/*@out@*/int*);
/* enthalpy function */
static double ufunct(double,long);
/* inverse enthalpy function */
static double inv_ufunct(double,long);


long NewGrainBin(void)
{
	long nd;

#	ifdef DEBUG_FUN
	fputs( "<+>NewGrainBin()\n", debug_fp );
#	endif

	if( gv.nBin >= NDUST ) 
	{
		fprintf( ioQQQ, " The code has run out of grain bins; increase NDUST and recompile.\n" );
		puts( "[Stop in NewGrainBin]" );
		cdEXIT(1);
	}
	nd = gv.nBin;
	if( gv.bin[nd] != NULL ) 
	{
		fprintf( ioQQQ, " Bin pointer not NULL, is memory already allocated ?\n" );
		puts( "[Stop in NewGrainBin]" );
		cdEXIT(1);
	}
	gv.bin[nd] = (GrainBin *)malloc(sizeof(GrainBin));
	if( gv.bin[nd] == NULL ) 
	{
		fprintf( ioQQQ, " Could not malloc new grain bin\n" );
		puts( "[Stop in NewGrainBin]" );
		cdEXIT(1);exit(1);
	}
	gv.lgDustOn = TRUE;
#	ifdef INCLUDE_OLD_GRAINS
	gv.bin[nd]->lgOldGrain1 = FALSE;
	gv.bin[nd]->lgDustOn1 = FALSE;
#	endif
	gv.bin[nd]->lgQHeat = FALSE;
	gv.bin[nd]->lgDustVary = FALSE;
	gv.bin[nd]->DustDftVel = FLT_MAX;
	gv.bin[nd]->avdft = 0.;
	gv.bin[nd]->TeGrainMax = FLT_MAX;
	gv.bin[nd]->dstfactor = 1.;
	/* this must be zero for the first solutions to be able to converge */
	/* >>chng 00 jun 19, tedust has to be greater than zero
	 * to prevent division by zero in GrainElecEmis and GrainCollHeating, PvH */
	gv.bin[nd]->tedust = 1.;
	/* used to check that energy scale in grains opacity files is same as
	 * current cloudy scale */
	gv.bin[nd]->EnergyCheck = 0.;
	gv.nBin++;

#	ifdef DEBUG_FUN
	fputs( " <->NewGrainBin()\n", debug_fp );
#	endif
	return(nd);
}

void ReturnGrainBins(void)
{
	long nd;

#	ifdef DEBUG_FUN
	fputs( "<+>ReturnGrainBins()\n", debug_fp );
#	endif

	for( nd=0; nd < gv.nBin; nd++ ) 
	{
		if( gv.bin[nd] == NULL ) 
		{
			fprintf( ioQQQ, " Bin pointer is NULL, is memory already freed ?\n" );
			puts( "[Stop in ReturnGrainBins]" );
			cdEXIT(1);
		}
		free( gv.bin[nd] );
		gv.bin[nd] = NULL;
	}
	gv.lgDustOn = FALSE;
	gv.nBin = 0;

#	ifdef DEBUG_FUN
	fputs( " <->ReturnGrainBins()\n", debug_fp );
#	endif
	return;
}


void InitGrains(void)
{
	long int i, 
	  nd,
	  nd2;
#	ifdef INCLUDE_OLD_GRAINS
	double InvDepCar[NCELL],
	  InvDepSil[NCELL];
#	endif

#	ifdef DEBUG_FUN
	fputs( "<+>InitGrains()\n", debug_fp );
#	endif

	if( trace.lgTrace && trace.lgDustBug )
	{
		fprintf( ioQQQ, " InitGrains called.\n" );
#		ifdef INCLUDE_OLD_GRAINS
		/* this is really only needed when old style grains are
		 * turned on, otherwise information below is enough... */
		fprintf( ioQQQ, "    There are %ld grain bins.\n", gv.nBin );
		fprintf( ioQQQ, "    Their names are:" );
		for( nd=0; nd < gv.nBin; nd++ ) 
		{
			fprintf( ioQQQ, "%s %c,", gv.bin[nd]->chDstLab, TorF(gv.bin[nd]->lgDustOn1) );
		}
		fprintf( ioQQQ, "\n" );
#		endif
	}

#	ifdef INCLUDE_OLD_GRAINS
	/* return memory of unused grain bins
	 * the first loop swaps all grain bins until
	 * the used ones are up front.
	 * the second loop frees the unused bins */
	for( nd=0; nd < gv.nBin; nd++ ) 
	{
		if( ! gv.bin[nd]->lgDustOn1 ) 
		{
			for( nd2 = nd+1; nd2 < gv.nBin; nd2++ ) 
			{
				if( gv.bin[nd2]->lgDustOn1 ) 
				{
					break;
				}
			}
			if( nd2 < gv.nBin ) 
			{
				GrainBin *swap;
				swap = gv.bin[nd];
				gv.bin[nd] = gv.bin[nd2];
				gv.bin[nd2] = swap;
			}
			else 
			{
				break;
			}
		}
	}
	for( nd2 = nd; nd2 < gv.nBin; nd2++ ) {
		if( gv.bin[nd2] != NULL ) {
			free( gv.bin[nd2] );
		}
		gv.bin[nd2] = NULL;
	}
	gv.nBin = nd;
	gv.lgDustOn = ( gv.nBin > 0 );
#endif

	/* sanity check */
	assert( gv.nBin >= 0 && gv.nBin < NDUST );
	for( nd=gv.nBin; nd < NDUST; nd++ ) 
	{
		assert( gv.bin[nd] == NULL );
	}

	for( i=0; i < rfield.nupper; i++ )
	{
		gv.dstab[i] = 0.;
		gv.dstsc[i] = 0.;
	}

	if( !gv.lgDustOn )
	{
		/* grains are not on, set all heating/cooling agents to zero */
		gv.GrainHeatInc = 0.;
		gv.GrainHeatDif = 0.;
		gv.GrainHeatLya = 0.;
		gv.GrainHeatCollSum = 0.;
		gv.GrainHeatSum = 0.;
		gv.GasCoolColl = 0.;
		HeatingCom.heating[13][0] = 0.;
		HeatingCom.heating[14][0] = 0.;
		HeatingCom.heating[25][0] = 0.;

		if( trace.lgTrace && trace.lgDustBug )
		{
			fprintf( ioQQQ, " InitGrains exits.\n" );
		}

#		ifdef DEBUG_FUN
		fputs( " <->InitGrains()\n", debug_fp );
#		endif
		return;
	}

	gv.lgAnyDustVary = FALSE;

#ifdef INCLUDE_OLD_GRAINS
	mie_read_ial("graphite.rfi",rfield.nupper,InvDepCar);
	mie_read_ial("silicate.rfi",rfield.nupper,InvDepSil);
#endif

	for( nd=0; nd < gv.nBin; nd++ )
	{
		/* sanity check */
		assert( gv.bin[nd] != NULL );

		/* this is QHEAT ALL command, will become default !! */
		if( gv.lgQHeatAll ) 
		{
			gv.bin[nd]->lgQHeat = TRUE;
		}

		/* this is NO GRAIN QHEAT command, always takes precedence */
		if( !gv.lgQHeatOn ) 
		{
			gv.bin[nd]->lgQHeat = FALSE;
		}

		gv.lgAnyDustVary = gv.lgAnyDustVary || gv.bin[nd]->lgDustVary;

		/* this is total grain depletion factor, GrainMetal is depletion
		 * set with metals command, normally one */
		gv.bin[nd]->dstfactor *= gv.GrainMetal;

		/* this will be actual abundance of grains used in calculation */
		gv.bin[nd]->dstAbund = (float)(gv.bin[nd]->dstfactor*GrnVryDpth(nd));
		assert( gv.bin[nd]->dstAbund > 0.f );

		if( gv.bin[nd]->DustWorkFcn < rfield.anu[0] || gv.bin[nd]->DustWorkFcn > rfield.anu[rfield.nupper] )
		{
			fprintf( ioQQQ, " Grain work function for %s has insane value: %.4e\n",
				 gv.bin[nd]->chDstLab,gv.bin[nd]->DustWorkFcn );
			puts( "[Stop in InitGrains]" );
			cdEXIT(1);
		}

#		ifdef INCLUDE_OLD_GRAINS
		if( gv.bin[nd]->lgOldGrain1 ) 
		{
			/* extrapolate absorption and scattering cross sections */
			OldStyleGrainCSInit(nd);

			/* initialize inverse attenuation length for photons */
			if( gv.bin[nd]->matType == MAT_CAR ) 
			{
				for( i=0; i < rfield.nupper; i++ ) 
				{
					gv.bin[nd]->inv_att_len[i] = (float)InvDepCar[i];
				}
			}
			else if( gv.bin[nd]->matType == MAT_SIL ) 
			{
				for( i=0; i < rfield.nupper; i++ ) 
				{
					gv.bin[nd]->inv_att_len[i] = (float)InvDepSil[i];
				}
			}
			else 
			{
				fprintf( ioQQQ, " InitGrains detected unknown material type: %ld\n" ,
					 gv.bin[nd]->matType );
				puts( "[Stop in InitGrains]" );
				cdEXIT(1);
			}
		}
#		endif

		for( i=0; i < rfield.nupper; i++ )
		{
			double alpha,beta,af,bf;

			beta = gv.bin[nd]->AvRadius*gv.bin[nd]->inv_att_len[i];
			if( beta > 1.e-4 ) 
			{
				bf = POW2(beta) - 2.*beta + 2. - 2.*exp(-beta);
			}
			else 
			{
				bf = POW3(beta)/3.;
			}
			alpha = beta + gv.bin[nd]->AvRadius/MEAN_PATH;
			if( alpha > 1.e-4 ) 
			{
				af = POW2(alpha) - 2.*alpha + 2. - 2.*exp(-alpha);
			}
			else 
			{
				af = POW3(alpha)/3.;
			}

			/* this is size-dependent geometrical yield enhancement
			 * defined in Weingartner & Draine, 2000 */
			gv.bin[nd]->y1[i] = (float)(POW2(beta/alpha)*af/bf);
		}

		/* this is effective area per proton, scaled by depletion
		 * dareff(nd) = darea(nd) * dstAbund(nd) */
		for( i=0; i < rfield.nupper; i++ )
		{
			/* these are total absorption and scattering cross sections */
			gv.dstab[i] += gv.bin[nd]->dstab1[i]*gv.bin[nd]->dstAbund;
			gv.dstsc[i] += gv.bin[nd]->dstsc1[i]*gv.bin[nd]->dstAbund;
		}
	}

	InitEmissivities();

	if( trace.lgDustBug && trace.lgTrace )
	{
		/*  >>chng 96 nov 30, two line format since soo many grains */
		/*  >>chng 00 may 11, even more grains -> yet more lines, PvH */
		fprintf( ioQQQ, "     There are %ld grain types turned on.\n", gv.nBin );

		fprintf( ioQQQ, "     grain depletion factors, dstfactor=" );
		for( nd=0; nd < gv.nBin; nd++ )
		{
			fprintf( ioQQQ, "%10.2e", gv.bin[nd]->dstfactor );
		}
		fprintf( ioQQQ, "\n" );

		fprintf( ioQQQ, "     lgDustVary flag for depth dependence:" );
		for( nd=0; nd < gv.nBin; nd++ )
		{
			fprintf( ioQQQ, "%2c", TorF(gv.bin[nd]->lgDustVary) );
		}
		fprintf( ioQQQ, " Any grain:%2c\n", TorF(gv.lgAnyDustVary) );

		fprintf( ioQQQ, "     Quantum heating flag:" );
		for( nd=0; nd < gv.nBin; nd++ )
		{
			fprintf( ioQQQ, "%2c", TorF(gv.bin[nd]->lgQHeat) );
		}
		fprintf( ioQQQ, "\n" );

		fprintf( ioQQQ, "     NU(Ryd), Abs cross sec per proton, total abs\n" );

		fprintf( ioQQQ, "    Ryd   " );
		for( nd=0; nd < gv.nBin; nd++ )
		{
			fprintf( ioQQQ, " %-12.12s", gv.bin[nd]->chDstLab );
		}
		fprintf( ioQQQ, " Total CS\n" );

		for( i=0; i < rfield.nupper; i += 40 )
		{
			fprintf( ioQQQ, "%10.2e", rfield.anu[i] );
			for( nd=0; nd < gv.nBin; nd++ )
			{
				fprintf( ioQQQ, " %10.2e  ", gv.bin[nd]->dstab1[i] );
			}
			fprintf( ioQQQ, "%10.2e\n", gv.dstab[i] );
		}

		fprintf( ioQQQ, "     NU(Ryd), Sct cross sec per proton, total sct\n" );

		fprintf( ioQQQ, "    Ryd   " );
		for( nd=0; nd < gv.nBin; nd++ )
		{
			fprintf( ioQQQ, " %-12.12s", gv.bin[nd]->chDstLab );
		}
		fprintf( ioQQQ, " Total CS\n" );

		for( i=0; i < rfield.nupper; i += 40 )
		{
			fprintf( ioQQQ, "%10.2e", rfield.anu[i] );
			for( nd=0; nd < gv.nBin; nd++ )
			{
				fprintf( ioQQQ, " %10.2e  ", gv.bin[nd]->dstsc1[i] );
			}
			fprintf( ioQQQ, "%10.2e\n", gv.dstsc[i] );
		}

		fprintf( ioQQQ, "     NU(Ryd), Q abs\n" );

		fprintf( ioQQQ, "    Ryd   " );
		for( nd=0; nd < gv.nBin; nd++ )
		{
			fprintf( ioQQQ, " %-12.12s", gv.bin[nd]->chDstLab );
		}
		fprintf( ioQQQ, "\n" );

		for( i=0; i < rfield.nupper; i += 40 )
		{
			fprintf( ioQQQ, "%10.2e", rfield.anu[i] );
			for( nd=0; nd < gv.nBin; nd++ )
			{
				fprintf( ioQQQ, " %10.2e  ", gv.bin[nd]->dstab1[i]*4./gv.bin[nd]->IntArea );
			}
			fprintf( ioQQQ, "\n" );
		}

		fprintf( ioQQQ, "     NU(Ryd), Q sct\n" );

		fprintf( ioQQQ, "    Ryd   " );
		for( nd=0; nd < gv.nBin; nd++ )
		{
			fprintf( ioQQQ, " %-12.12s", gv.bin[nd]->chDstLab );
		}
		fprintf( ioQQQ, "\n" );

		for( i=0; i < rfield.nupper; i += 40 )
		{
			fprintf( ioQQQ, "%10.2e", rfield.anu[i] );
			for( nd=0; nd < gv.nBin; nd++ )
			{
				fprintf( ioQQQ, " %10.2e  ", gv.bin[nd]->dstsc1[i]*4./gv.bin[nd]->IntArea );
			}
			fprintf( ioQQQ, "\n" );
		}

		fprintf( ioQQQ, "     NU(Ryd), yield enhancement\n" );

		fprintf( ioQQQ, "    Ryd   " );
		for( nd=0; nd < gv.nBin; nd++ )
		{
			fprintf( ioQQQ, " %-12.12s", gv.bin[nd]->chDstLab );
		}
		fprintf( ioQQQ, "\n" );

		for( i=0; i < rfield.nupper; i += 40 )
		{
			fprintf( ioQQQ, "%10.2e", rfield.anu[i] );
			for( nd=0; nd < gv.nBin; nd++ )
			{
				fprintf( ioQQQ, " %10.2e  ", gv.bin[nd]->y1[i] );
			}
			fprintf( ioQQQ, "\n" );
		}

		fprintf( ioQQQ, " InitGrains exits.\n" );
	}


#	ifdef DEBUG_FUN
	fputs( " <->InitGrains()\n", debug_fp );
#	endif
	return;
}


static void InitEmissivities()
{
	double fac,
	  fac2,
	  mul,
	  tdust;
	long int i,
	  nd;

#	ifdef DEBUG_FUN
	fputs( "<+>InitEmissivities()\n", debug_fp );
#	endif

	if( trace.lgTrace && trace.lgDustBug )
	{
		fprintf( ioQQQ, "  InitEmissivities starts\n" );
		fprintf( ioQQQ, "    ND    Tdust       Emis       BB Check   4pi*a^2*<Q>\n" );
	}

	assert( NDEMS > 2*NTOP );
	fac = exp(log(GRAIN_TMID/GRAIN_TMIN)/(double)(NDEMS-NTOP));
	tdust = GRAIN_TMIN;
	for( i=0; i < NDEMS-NTOP; i++ )
	{
		gv.dsttmp[i] = log(tdust);
		for( nd=0; nd < gv.nBin; nd++ )
		{
			gv.bin[nd]->dstems[i] = log(PlanckIntegral(tdust,nd,i));
		}
		tdust *= fac;
	}

	/* temperatures above GRAIN_TMID are unrealistic -> make grid gradually coarser */
	fac2 = exp(log(GRAIN_TMAX/GRAIN_TMID/powi(fac,NTOP-1))/(double)((NTOP-1)*NTOP/2));
	for( i=NDEMS-NTOP; i < NDEMS; i++ )
	{
		gv.dsttmp[i] = log(tdust);
		for( nd=0; nd < gv.nBin; nd++ )
		{
			gv.bin[nd]->dstems[i] = log(PlanckIntegral(tdust,nd,i));
		}
		fac *= fac2;
		tdust *= fac;
	}

	/* sanity checks */
	mul = 1.;
	assert( fabs(gv.dsttmp[0] - log(GRAIN_TMIN)) < 10.*mul*DBL_EPSILON );
	mul = sqrt((double)(NDEMS-NTOP));
	assert( fabs(gv.dsttmp[NDEMS-NTOP] - log(GRAIN_TMID)) < 10.*mul*DBL_EPSILON );
	mul = (double)NTOP + sqrt((double)NDEMS);
	assert( fabs(gv.dsttmp[NDEMS-1] - log(GRAIN_TMAX)) < 10.*mul*DBL_EPSILON );

	/* now find slopes form spline fit */
	for( nd=0; nd < gv.nBin; nd++ )
	{
		/* set up coefficients for spline */
		spline(gv.bin[nd]->dstems,gv.dsttmp,NDEMS,2e31,2e31,gv.bin[nd]->dstslp);
		spline(gv.dsttmp,gv.bin[nd]->dstems,NDEMS,2e31,2e31,gv.bin[nd]->dstslp2);
	}

#	ifdef DEBUG_FUN
	fputs( " <->InitEmissivities()\n", debug_fp );
#	endif
	return;
}

/* PlanckIntegral compute total radiative cooling due to grains */
static double PlanckIntegral(double tdust, 
			     long int nd, 
			     long int ip)
{
	long int i;

	double arg,
	  ExpM1,
	  integral1 = 0.,  /* integral(Planck) */
	  integral2 = 0.,  /* integral(Planck*abs_cs) */
	  Planck1,
	  Planck2,
	  TDustRyg, 
	  x;

#	ifdef DEBUG_FUN
	fputs( "<+>PlanckIntegral()\n", debug_fp );
#	endif

	/******************************************************************
	 *
	 * >>>chng 99 mar 12, this sub rewritten following Peter van Hoof
	 * comments.  Original coding was in single precision, and for
	 * very low temperature the exponential was set to zero.  As 
	 * a result Q was far too large for grain temperatures below 10K
	 *
	 ******************************************************************/

	/* Boltzmann factors for Planck integration */
	TDustRyg = TE1RYD/tdust;

	x = 0.999*log(DBL_MAX);

	for( i=0; i < rfield.nupper; i++ )
	{
		/* this is hnu/kT for grain at this temp and photon energy */
		arg = TDustRyg*rfield.anu[i];

		/* want the number exp(hnu/kT) - 1, two expansions */
		if( arg < 1.e-5 )
		{
			/* for small arg expand exp(hnu/kT) - 1 to second order */
			ExpM1 = arg*(1. + arg/2.);
		}
		else
		{
			/* for large arg, evaluate the full expansion */
			ExpM1 = exp(MIN2(x,arg)) - 1.;
		}

		Planck1 = PI4*2.*HPLANCK/POW2(SPEEDLIGHT)*POW2(FR1RYD)*POW2(FR1RYD)*
			rfield.anu3[i]/ExpM1*rfield.widflx[i];
		Planck2 = Planck1*gv.bin[nd]->dstab1[i];

		/* add integral over RJ tail, maybe useful for extreme low temps */
		if( i == 0 ) 
		{
			integral1 = Planck1/rfield.widflx[0]*rfield.anu[0]/3.;
			integral2 = Planck2/rfield.widflx[0]*rfield.anu[0]/5.;
		}
		/* if we are in the Wien tail - exit */
		if( Planck1/integral1 < DBL_EPSILON && Planck2/integral2 < DBL_EPSILON ) {
			break;
		}
		integral1 += Planck1;
		integral2 += Planck2;
	}

	/* this is an option to print out every few steps, when 'trace grains' is set */
	if( trace.lgDustBug && trace.lgTrace && ip%10 == 0 )
	{
		fprintf( ioQQQ, "  %4ld %11.4e %11.4e %11.4e %11.4e\n", nd, tdust, 
		  integral2, integral1/4./5.67051e-5/powi(tdust,4), integral2*4./integral1 );
	}

	assert( integral2 > 0. );

#	ifdef DEBUG_FUN
	fputs( " <->PlanckIntegral()\n", debug_fp );
#	endif
	return( integral2 );
}


/* this is the main routine that drives the grain physics */
void grain(void)
{
	long int i,
	  ipZ,
	  nd;

#	ifdef DEBUG_FUN
	fputs( "<+>grain()\n", debug_fp );
#	endif

	if( gv.lgDustOn )
	{		
		/* option to have grain abundance depend on depth */
		if( gv.lgAnyDustVary )
		{
			for( i=0; i < rfield.nupper; i++ )
			{
				/* these are total absorption and scattering cross sections */
				gv.dstab[i] = 0.;
				gv.dstsc[i] = 0.;
			}

			for( nd=0; nd < gv.nBin; nd++ )
			{
				/* refresh grain opacities if variable grain abundances enabled */
				if( gv.bin[nd]->lgDustVary )
				{
					gv.bin[nd]->dstAbund = (float)(gv.bin[nd]->dstfactor*GrnVryDpth(nd));
					assert( gv.bin[nd]->dstAbund > 0.f );
				}

				for( i=0; i < rfield.nupper; i++ )
				{
					/* these are total absorption and scattering cross sections */
					gv.dstab[i] += gv.bin[nd]->dstab1[i]*gv.bin[nd]->dstAbund;
					gv.dstsc[i] += gv.bin[nd]->dstsc1[i]*gv.bin[nd]->dstAbund;
				}
			}
		}

		for( ipZ=0; ipZ < LIMELM; ipZ++ ) 
		{
			/* this is average (NOT rms) particle speed for Maxwell distribution, Mihalas 70, 9-70 */
			gv.AveVel[ipZ] = sqrt(8.*BOLTZMANN/PI/ATOMIC_MASS_UNIT*phycon.te/AtomcWgt.AtomicWeight[ipZ]);
		}

		/* find dust charge and temperature */
		GrainChargeTemp();

		/* find dust drift velocity */
		GrainDrift();
	}

#	ifdef DEBUG_FUN
	fputs( " <->grain()\n", debug_fp );
#	endif
	return;
}

/* iterate grain charge and temperature */
static void GrainChargeTemp(void)
{
	long int nd;
	float dccool;
	double delta,
	  dtherm,
	  hcon,
	  hla,
	  hots,
	  oldtemp,
	  ratio,
	  thermionic,
	  ThermRatio;

#	ifdef DEBUG_FUN
	fputs( "<+>GrainChargeTemp()\n", debug_fp );
#	endif

	if( trace.lgTrace && trace.lgDustBug )
	{
		fprintf( ioQQQ, "    GrainChargeTemp called lgSearch%2c\n", TorF(conv.lgSearch) );
	}

	/* these will sum heating agents over grain populations */
	gv.GrainHeatInc = 0.;
	gv.GrainHeatDif = 0.;
	gv.GrainHeatLya = 0.;
	gv.GrainHeatCollSum = 0.;
	gv.GrainHeatSum = 0.;

	gv.GasCoolColl = 0.;
	/* must redo grain solution all the time since they
	 * generally have non-trivial contributions to heating-cooling */
	HeatingCom.heating[13][0] = 0.;
	HeatingCom.heating[25][0] = 0.;

	gv.TotalEden = 0.;

	for( nd=0; nd < gv.nBin; nd++ )
	{
		if( trace.lgTrace && trace.lgDustBug )
		{
			fprintf( ioQQQ, "    >> GrainChargeTemp starting grain %s\n",
				 gv.bin[nd]->chDstLab );
		}

		do {
			oldtemp = gv.bin[nd]->tedust;

			/* solve for charge using previous estimate for grain temp
			 * grain temp only influences thermionic emissions
			 * Thermratio is fraction thermionic emissions contribute
			 * to the total electron loss rate of the grain */
			GrainCharge(nd,&ThermRatio);

			/* now solve grain temp using new value for grain potential */
			GrainTemperature(nd,&dccool,&hcon,&hots,&hla,&dtherm,&thermionic);

			/* delta estimates relative change in electron emission rate
			 * due to the update in the grain temperature, if it is small
			 * we won't bother to iterate (which is usually the case)
			 * the formula assumes that thermionic emission is the only
			 * process that depends on grain temperature */
			ratio = gv.bin[nd]->tedust/oldtemp;
			delta = gv.bin[nd]->ThresInf*TE1RYD/gv.bin[nd]->tedust*(ratio - 1.);
			delta = ( delta < 0.9*log(DBL_MAX) ) ?
				ThermRatio*fabs(POW2(ratio)*exp(delta)-1.) : DBL_MAX;

			if( trace.lgTrace && trace.lgDustBug )
			{
				fprintf( ioQQQ, "    >> GrainChargeTemp finds delta=%.4e, ",delta );
				fprintf( ioQQQ, " old/new temp=%.4e %.4e, ",oldtemp,gv.bin[nd]->tedust );
				if( delta > TOLER ) 
				{
					fprintf( ioQQQ, "doing another iteration\n" );
				}
				else 
				{
					fprintf( ioQQQ, "converged\n\n" );
				}
			}
		} while( delta > TOLER );

		/* following used to keep track of heating agents in printout
		 * no physics done with GrainHeatInc
		 * dust heating by incident continuum, and elec friction before ejection */
		gv.GrainHeatInc += hcon + dtherm;
		/* remember total heating by diffuse fields, for printout (includes Lya) */
		gv.GrainHeatDif += hots;
		/* GrainHeatLya - total heating by LA in this zone, erg cm-3 s-1, only here
		 * for eventual printout, hots is total ots line heating */
		gv.GrainHeatLya += hla;

		/* this will be total collisional heating, for printing in lines */
		gv.GrainHeatCollSum += gv.bin[nd]->GrainHeatColl;

		/* GrainHeatSum is total heating of all grain types in this zone,
		 * will be carried by total cooling, only used in lines to print tot heat
		 * printed as entry "GraT    0 " */
		gv.GrainHeatSum += gv.bin[nd]->GrainHeat;

		/* dccool is gas cooling due to collisions with grains */
		gv.GasCoolColl += dccool;

		/* rate dust heats gas by photoelectric effect */
		HeatingCom.heating[13][0] += gv.bin[nd]->GasHeatPhotoEl;
		/* thermionic is gas heating due to thermionic emissions */
		HeatingCom.heating[25][0] += thermionic;

		/* this is grain charge in e/cm^3, positive number means grain supplied free electrons */
		gv.TotalEden += gv.bin[nd]->DustZ*phycon.hden*gv.bin[nd]->dstAbund/gv.bin[nd]->qhnrat;

		if( trace.lgTrace && trace.lgDustBug )
		{
			fprintf(ioQQQ," GrainChargeTemp %s Pot %.5e GasPEHeat %.5e GasCoolColl %.5e\n" , 
				gv.bin[nd]->chDstLab, gv.bin[nd]->dstpot, gv.bin[nd]->GasHeatPhotoEl, dccool );
		}
	}

	/* if grains hotter than gas then collisions with gas act
	 * to heat the gas, add this in here
	 * a symmetric statement appears in COOLR, where cooling is added on */
	HeatingCom.heating[14][0] = MAX2(0.,-gv.GasCoolColl);

	/* options to force gas heating or cooling by grains to zero - for tests only ! */
	if( !gv.lgDHetOn ) 
	{
		HeatingCom.heating[13][0] = 0.;
		HeatingCom.heating[14][0] = 0.;
		HeatingCom.heating[25][0] = 0.;
	}
	if( !gv.lgDColOn ) 
	{
		gv.GasCoolColl = 0.;
	}

	if( trace.lgTrace )
	{
		if( trace.lgDustBug )
		{
			fprintf( ioQQQ, "     Grain contribution to electron density%10.2e\n", gv.TotalEden );
		}
		else
		{
			fprintf( ioQQQ, "     Grain potentials:" );
			for( nd=0; nd < gv.nBin; nd++ )
			{
				fprintf( ioQQQ, "%10.2e", gv.bin[nd]->dstpot );
			}
			fprintf( ioQQQ, "\n" );
			
			fprintf( ioQQQ, "     Grain temperatures:" );
			for( nd=0; nd < gv.nBin; nd++ )
			{
				fprintf( ioQQQ, "%10.2e", gv.bin[nd]->tedust );
			}
			fprintf( ioQQQ, "\n" );
		}
	}


#	ifdef DEBUG_FUN
	fputs( " <->GrainChargeTemp()\n", debug_fp );
#	endif
	return;
}

static void GrainCharge(long int nd,
			double *ThermRatio) /* ratio of thermionic to total rate */
{
	int lgBigChange=FALSE,
	  lgBigError,
	  lgGoodProgress,
	  lgNR;
	long int lopbig,
	  luplim;
	float maxlo,
	  minhi;
	double change,
	  dnew,
	  emission = -DBL_MAX,
	  fhi,
	  flo,
	  fmid, 
	  max_slope,
	  old_f,
	  old_pot,
	  old_slope,
	  recombination = -DBL_MAX,
	  rerror,
	  slope,
	  xxx;

#	ifdef DEBUG_FUN
	fputs( "<+>GrainCharge()\n", debug_fp );
#	endif

	if( conv.lgSearch )
	{
		/* this is limit to number of loops we will try before giving up on getting
		 * the grain charge right, set to 100 when in search phase, 10 otherwise */
		luplim = 100;

		/* set up initial estimates for the grain potential */
		gv.bin[nd]->dstpot = 0.;
		/* update items that depend on grain potential */
		UpdatePot(nd);
	}
	else
	{
		/* >>chng 00 jul 19, convergence around zero pot can be tricky; luplim 10 -> 20, PvH */
		luplim = 20;
	}

	if( trace.lgTrace && trace.lgDustBug )
	{
		fprintf( ioQQQ, "    Starting Charge loop for %s, search%2c\n",
			 gv.bin[nd]->chDstLab, TorF(conv.lgSearch) );
	}

	/* find dust charge */
	lgBigError = TRUE;
	/* grain potential may never be lower than this */
	maxlo = (gv.bin[nd]->BandGap-gv.bin[nd]->DustWorkFcn)/gv.bin[nd]->TunnelCoef;
	/* arbitrary first estimate for upper bracket
	 * NB - current value of gv.bin[nd]->dstpot must be inside bracket ! */
	minhi = MAX2(BIG_POTENTIAL,gv.bin[nd]->dstpot+0.5f*BIG_POTENTIAL);

	/* >>chng 01 jan 08, validate bracket and alter if
	 * necessary, set first step in gv.bin[nd]->dstpot 
	 * (usually the value from the previous zone) */
	BracketGrainPot(nd,&maxlo,&flo,&minhi,&fhi,ThermRatio);

	/* if BracketGrainPot accidentally stumbled on the
	 * correct solution, it will be in gv.bin[nd]->dstpot */
	old_pot = minhi;
	old_f = fhi;
	old_slope = (fhi-flo)/(double)(minhi-maxlo);
	slope = old_slope;

	/* sanity check */
	assert( slope < 0. );

	for( lopbig=0; lopbig < luplim && lgBigError; lopbig++ )
	{
		/* function GrainElecEmis returns electron loss rate for the grain species
		 * due to photoelectric effect, positive ion recombinations on the
		 * grain surface and thermionic emissions */
		emission = GrainElecEmis(nd,ThermRatio);

		/* the function GrainElecRecomb returns the recombination rate
		 * for this grain and charge */
		recombination = GrainElecRecomb(nd);

		/* this is the difference between emission and recombination
		 * rates - we want this to be small */
		fmid = emission - recombination;

		/* update bracket for the correct solution */
		if( fmid > 0. ) 
		{
			/* current potential is below correct value */
			maxlo = MAX2(maxlo,gv.bin[nd]->dstpot);
		}
		else 
		{
			/* current potential is above correct value */
			minhi = MIN2(minhi,gv.bin[nd]->dstpot);
		}

		/* >>chng 97 feb 28, crashed when both emission and recombination zero -
		 * which happens when no ionizing radiation */
		/* >>chng 99 jun 16, grainlte.in has no ionizing rad at all, and loop
		 * never converged.  add second test for convergence, since first test
		 * always fails when emission is zero and recombination is relative to
		 * recombination */
		/* >>chng 00 jun 21, the inclusion of thermionic emissions has solved
		 * the above problem and grainlte.in now converges properly, hence
		 * RecNeutral is not needed anymore, PvH */
		if( emission+recombination > SMALLFLOAT )
		{
			rerror = fabs(fmid) / (0.5*(emission+recombination));
		}
		else
		{
			/* this clause implies that there is no ionizing radiation and
			 * also that the grains are cold enough that thermionic emissions
			 * are unimportant. the first constraint implies that grain surface
			 * recombination rates for electrons and positive particles are
			 * negligible, while the second constraint implies that the
			 * photoelectric effect is unimportant. in all likelyhood balmer
			 * photons are the only source of heating for the grains and therefore
			 * the heating does not depend on grain potential, which is good since
			 * we cannot determine it -> set rerror=0 to fake proper convergence, PvH */
			rerror = 0.;
		}

		if( rerror > TOLER )
		{
			if( gv.bin[nd]->dstpot - old_pot != 0. )
			{
				/*   only re-estimate slope if new continuum
				 * >>chng 96 apr 26, as per Peter van Hoof comment
				 * that div by zero can occur */
				xxx = (fmid-old_f)/(gv.bin[nd]->dstpot - old_pot);
				/* >>chng 00 aug 10, slope should always be negative, if it is not
				 * it indicates problems due to a change in electron temp, by PvH */
				if( xxx < 0. )
					slope = xxx;
			}

			/* big changes in the slope could indicate oscilating behaviour, PvH */
			max_slope = MAX2(fabs(old_slope),fabs(slope));
			if( lopbig > 2 )
			{
				double min_slope = MIN2(fabs(old_slope),fabs(slope));
				lgBigChange = max_slope/min_slope > BIG_CHANGE;
				{
					/*@-redef@*/
					enum {DEBUG=FALSE};
					/*@+redef@*/
					if( DEBUG && lgBigChange ) {
						long i;
						int lgBackup = trace.lgDustBug;
						float backup = gv.bin[nd]->dstpot;
						/* switch off trace output temporarily */
						trace.lgDustBug = FALSE;
						for( i=0; i<=250; i++ ) {
							gv.bin[nd]->dstpot = maxlo + (float)i/250.f*(minhi-maxlo);
							UpdatePot(nd);
							emission = GrainElecEmis(nd,ThermRatio);
							recombination = GrainElecRecomb(nd);
							printf(" BigChangeTest: %.6e %.6e %.6e\n",
							       gv.bin[nd]->dstpot,emission,recombination );
						}
						/* reset to old values */
						gv.bin[nd]->dstpot = backup;
						UpdatePot(nd);
						emission = GrainElecEmis(nd,ThermRatio);
						recombination = GrainElecRecomb(nd);
						trace.lgDustBug = lgBackup;
					}
				}
			}
			else
			{
				lgBigChange = FALSE;
			}

			/* force Newton-Raphson until slope has settled in */
			lgGoodProgress = ( lopbig > 0 ) ? fabs(old_f/fmid) > GOOD_PROGRESS : TRUE;

			old_pot = gv.bin[nd]->dstpot;
			old_f = fmid;
			old_slope = slope;

			/* here is reason for above xxx, do not want zero slope */
			dnew = -fmid/slope;
			change = fabs(dnew);
			/* >>chng 00 jul 14, make sure change is not infinitesimally small, PvH */
			change = MAX2(change,(double)(DMINX*(minhi-maxlo)));
			/* >>chng 01 jan 08, removed upper limit on change, code below will handle this, PvH */
			change = sign(change,dnew);

			gv.bin[nd]->dstpot += (float)change;
			lgNR = TRUE;

			/* >>chng 00 jun 19, make sure grain potential never drops below
			 * bandgap minus workfunction plus potsurf, electrons cannot be bound in 
			 * that case, it also prevents FP overflow in the thermionic rate, PvH */
			/* >>chng 00 jul 18, keep solution bracketed, this is needed since
			 * bandgap introduces discontinuities in emission and recombination rate;
			 * solution can keep oscillating around zero, PvH */
			/* >>chng 01 jan 03, lgBigChange tracks huge jumps in the slope, which can
			 * indicate oscillating behaviour of the code, PvH */
			/* >>chng 01 jan 08, lgGoodProgess indicates whether fmid decreases fast
			 * enough. if it does not use bisection to speed up convergence, PvH */
			if( lgBigChange || ! lgGoodProgress ||
			    gv.bin[nd]->dstpot <= maxlo || gv.bin[nd]->dstpot >= minhi ) 
			{
				gv.bin[nd]->dstpot = (maxlo + minhi)/2.f;
				lgNR = FALSE;
			}
				 
			/* update items that depend on grain potential */
			UpdatePot(nd);

			/* if minhi and maxlo are close enough we declare convergence */
			lgBigError = ( max_slope*(minhi-maxlo) > TOLER*0.5*(emission+recombination) ||
				       minhi-maxlo > DELTA_BANDGAP );
		}
		else
		{
			old_pot = gv.bin[nd]->dstpot;
			lgNR = FALSE;
			lgBigError = FALSE;
		}

		if( trace.lgTrace && trace.lgDustBug )
		{
			fprintf( ioQQQ, "  Charge loop %ld %s used pot %.2e next pot %.2e Vlo %.2e", 
			  lopbig, gv.bin[nd]->chDstLab, old_pot, gv.bin[nd]->dstpot, maxlo );
			fprintf( ioQQQ, " Vhi %.2e BC%2c NR%2c tol %.2e slp %.2e Em/Rc %.2e %.2e\n", 
			  minhi, TorF(lgBigChange), TorF(lgNR), rerror, slope, emission, recombination );
		}
	}

	if( lgBigError && !conv.lgSearch )
	{
		fprintf( ioQQQ," GrainCharge did not converge ionization of grain species %s\n",
			 gv.bin[nd]->chDstLab );
		++conv.nGrainFail;
	}

	if( trace.lgTrace && trace.lgDustBug )
	{
		fprintf( ioQQQ, " Grain potential:%12.12s%10.2eRyd, emis:%10.2e recom:%10.2e", 
		  gv.bin[nd]->chDstLab, gv.bin[nd]->dstpot, emission, recombination );
		fprintf( ioQQQ, " Threshold:%10.2eRyd\n", 
		  gv.bin[nd]->ThresInf );
	}


#	ifdef DEBUG_FUN
	fputs( " <->GrainCharge()\n", debug_fp );
#	endif
	return;
}

/* bracket the solution for the grain potential */
static void BracketGrainPot(long nd,
			    /*@in@*/float* xlo,
			    /*@out@*/double* flo,
			    /*@in@*/float* xhi,
			    /*@out@*/double* fhi,
			    /*@out@*/double* ThermRatio)
{
	long loop=0;
	float estimate;

#	ifdef DEBUG_FUN
	fputs( "<+>BracketGrainPot()\n", debug_fp );
#	endif

	estimate = gv.bin[nd]->dstpot;

	gv.bin[nd]->dstpot = *xlo;
	UpdatePot(nd);
	*flo = GrainElecEmis(nd,ThermRatio) - GrainElecRecomb(nd);
	if( *flo < 0. )
	{
		fprintf( ioQQQ, " insanity: grain potential below lowest allowed value for %s\n" ,
			 gv.bin[nd]->chDstLab );
		ShowMe();
		puts( "[Stop in BracketGrainPot]" );
		cdEXIT(1);
	}

	*fhi = 1.;
	while( *fhi > 0. && loop < BRACKET_MAX )
	{
		gv.bin[nd]->dstpot = *xhi;
		UpdatePot(nd);
		*fhi = GrainElecEmis(nd,ThermRatio) - GrainElecRecomb(nd);
		if( *fhi > 0. )
		{
			*xlo = *xhi;
			*flo = *fhi;
			*xhi += 5.f;
			/* value from previous zone is useless now */
			estimate = (*xlo + *xhi)/2.f;
		}
		loop++;
	}
	if( *fhi > 0. )
	{
		fprintf( ioQQQ, " could not bracket grain potential for %s\n" , gv.bin[nd]->chDstLab );
		fprintf( ioQQQ, " please increase BRACKET_MAX, current value is %d\n", BRACKET_MAX );
		puts( "[Stop in BracketGrainPot]" );
		cdEXIT(1);
	}

	/* set initial estimate for grain potential,
	 * usually the value from the previous zone,
	 * unless we accidentally hit exact solution */
	if( *flo == 0. )
		estimate = *xlo;
	if( *fhi == 0. )
		estimate = *xhi;
	gv.bin[nd]->dstpot = estimate;
	UpdatePot(nd);

#	ifdef DEBUG_FUN
	fputs( " <->BracketGrainPot()\n", debug_fp );
#	endif
	return;
}

static double GrainElecRecomb(long int nd)
{
	long ion,
	  ipZ;
	double atoms,
	  eta,
	  p_rad,
	  rate,
	  Stick,
	  sum1,
	  sum2,
	  ve,
	  xi;

#	ifdef DEBUG_FUN
	fputs( "<+>GrainElecRecomb()\n", debug_fp );
#	endif

	/* -1 makes psi correct for impact by electrons */
	ion = -1;
	/* >>chng 00 jul 19, replace classical results with results including image potential
	 * to correct for polarization of the grain as charged particle approaches. */
	GrainScreen(nd,ion,&eta,&xi);

	/* VE is mean (not RMS) electron velocity */
	/*ve = phycon.sqrte*6.2124e5;*/
	ve = sqrt(8.*BOLTZMANN/PI/ELECTRON_MASS*phycon.te);

	/* >>chng 00 jul 13, new sticking probability for electrons */
	if( gv.bin[nd]->dstpot <= 0. ) 
	{
		atoms = gv.bin[nd]->AvVol*gv.bin[nd]->dustp[0]/ATOMIC_MASS_UNIT/gv.bin[nd]->atomWeight;
		p_rad = 1./(1.+exp(20.-atoms));
	}
	else 
	{
		p_rad = 1.;
	}
	/* the second term is chance that electron passes through grain,
	 * 1-p_rad is chance that electron is ejected before grain settles
	 * see discussion in Weingartner & Draine, 2000, ApJ ....  */
	Stick = STICK_ELEC*(1. - exp(-gv.bin[nd]->AvRadius/MEAN_PATH))*p_rad;
	/* this is grain surface recomb rate for electrons */
	sum1 = Stick*phycon.eden*ve*eta;

	/* >>chng 00 jul 13, add in gain rate from atoms and ions, PvH */
	sum2 = 0.;
	for( ipZ=0; ipZ < LIMELM; ipZ++ )
	{
		if( abundances.lgElmtOn[ipZ])
		{		
			for( ion=0; ion < gv.bin[nd]->RecomZ0[ipZ]; ion++ )
			{
				/* >>chng 00 jul 19, replace classical results with results
				 * including image potential to correct for polarization 
				 * of the grain as charged particle approaches. */
				GrainScreen(nd,ion,&eta,&xi);
				/* this is rate with which charged ion strikes grain */
				sum2 += STICK_ION*xIonFracs[ion+1][ipZ]*gv.AveVel[ipZ]*
					(double)(gv.bin[nd]->RecomZ0[ipZ]-ion)*eta;
			}
		}
	}

	rate = sum1 + sum2;

	if( trace.lgTrace && trace.lgDustBug )
	{
		fprintf( ioQQQ, "    GrainElecRecomb finds rate1=%.4e, rate2=%.4e.\n",sum1,sum2 );
		if( rate > 0. ) 
		{
			fprintf( ioQQQ, "    rel. fractions rate1/sum=%.4e, rate2/sum=%.4e.\n",
				 sum1/rate,sum2/rate );
		}
	}


#	ifdef DEBUG_FUN
	fputs( " <->GrainElecRecomb()\n", debug_fp );
#	endif
	return( rate );
}

/* GrainElecEmis computes electron loss rate for this grain species due
 * to photoelectric effect, positive ion recombinations on the
 * grain surface and thermionic emissions */
static double GrainElecEmis(long int nd,
			    double *ThermRatio) /* ratio of thermionic to total emissions */
{
	long int i,
	  ion,
	  ipZ;
	double eta,
	  sum,
	  sum1,
	  sum2,
	  sum3,
	  ThermExp,
	  xi;

#	ifdef DEBUG_FUN
	fputs( "<+>GrainElecEmis()\n", debug_fp );
#	endif

	/* this is the loss rate due to photo-electric effect */
	sum1 = 0.;
	for( i=gv.bin[nd]->ipDustPot-1; i < rfield.nflux; i++ )
	{
		/* >>chng 00 jul 17, use description of Weingartner & Draine, 2000 */
		sum1 += rfield.SummedCon[i]*gv.bin[nd]->yhat[i]*gv.bin[nd]->dstab1[i];
	}
	sum1 /= gv.bin[nd]->IntArea/4.;

	/* >>chng 00 jun 19, add in loss rate due to recombinations with ions, PvH */
	sum2 = 0.;
	for( ion=0; ion <= LIMELM; ion++ )
	{
		double CollisionRateAll;

		/* >>chng 00 jul 19, replace classical results with results including image potential
		 * to correct for polarization of the grain as charged particle approaches. */
		GrainScreen(nd,ion,&eta,&xi);

		CollisionRateAll = 0.;

		for( ipZ=MAX2(ion-1,0); ipZ < LIMELM; ipZ++ )
		{
			if( ion > gv.bin[nd]->RecomZ0[ipZ] )
			{
				/* this is rate with which charged ion strikes grain */
				CollisionRateAll += STICK_ION*xIonFracs[ion+1][ipZ]*gv.AveVel[ipZ]*
					(double)(ion-gv.bin[nd]->RecomZ0[ipZ]);
			}
		}

		sum2 += CollisionRateAll*eta;
	}

	/* >>chng 00 jun 19, add in loss rate due to thermionic emission of electrons, PvH */
	ThermExp = gv.bin[nd]->ThresInf*TE1RYD/gv.bin[nd]->tedust;
	/* ThermExp is guaranteed to be >= 0. */
	gv.bin[nd]->ThermRate = THERMCONST*gv.bin[nd]->ThermEff*POW2(gv.bin[nd]->tedust)*exp(-ThermExp);
	sum3 = gv.bin[nd]->ThermRate;

	sum = sum1+sum2+sum3;

	*ThermRatio = ( sum > 0. ) ? sum3/sum : 0.;
		
	if( trace.lgTrace && trace.lgDustBug )
	{
		fprintf( ioQQQ, "    GrainElecEmis finds rate1=%.4e, rate2=%.4e, rate3=%.4e.\n",sum1,sum2,sum3 );
		if( sum > 0. ) 
		{
			fprintf( ioQQQ, "    rel. fractions rate1/sum=%.4e, rate2/sum=%.4e, rate3/sum=%.4e.\n",
				 sum1/sum,sum2/sum,sum3/sum );
		}
	}

#	ifdef DEBUG_FUN
	fputs( " <->GrainElecEmis()\n", debug_fp );
#	endif
	return( sum );
}

/* correction factors for grain charge screening (including image potential
 * to correct for polarization of the grain as charged particle approaches). */
static void GrainScreen(long nd,
			long ion,
			/*@out@*/ double *eta,
			/*@out@*/ double *xi)
{
	long Z;
	double frac,
	  eta1,
	  xi1,
	  sum,
	  nu,
	  tau,
	  theta_nu,
	  xi_nu,
	  xxx;

#	ifdef DEBUG_FUN
	fputs( "<+>GrainScreen()\n", debug_fp );
#	endif

	/* >>ref Draine & Sutin, 1987, ApJ, 320, 803
	 * eta = J-tilde (eq. 3.3 thru 3.5), xi = Lambda-tilde/2. (eq. 3.8 thru 3.10) */
	if( ion == 0 ) 
	{
		*eta = 1.;
		*xi = 1.;
	}
	else 
	{
		sum = 0.;
		*eta = 0.;
		*xi = 0.;
		/* >>chng 01 jan 03, assume that grain charge distribution has a 1-sigma width SIG_DUSTZ
		 * instead of the delta function we assume elsewhere. by averaging over the distribution we
		 * smooth out the discontinuities of the the Draine & Sutin expressions near nu == 0. they
		 * were the cause of temperature instabilities in the globule.in test case. PvH
		 * NB - the grain charge distribution should be calculated self-consistently of course ! */
		for( Z = NINT(gv.bin[nd]->DustZ-3.*SIG_DUSTZ); Z <= NINT(gv.bin[nd]->DustZ+3.*SIG_DUSTZ); Z++ )
		{
			frac = exp(-POW2(((double)Z-gv.bin[nd]->DustZ)/SIG_DUSTZ));
			nu = (double)Z/(double)ion;
			tau = gv.bin[nd]->Capacity*BOLTZMANN*phycon.te*1.e-7/POW2((double)ion*ELEM_CHARGE);
			if( nu < 0. )
			{
				eta1 = (1. - nu/tau)*(1. + sqrt(2./(tau - 2.*nu)));
				xi1 = (1. - nu/(2.*tau))*(1. + 1./sqrt(tau - nu));
			}
			else if( nu == 0. ) 
			{
				eta1 = 1. + sqrt(PI/(2.*tau));
				xi1 = 1. + 0.75*sqrt(PI/(2.*tau));
			}
			else 
			{
				xi_nu = 1. + 1./sqrt(3.*nu);
				theta_nu = nu/xi_nu - 1./(2.*POW2(xi_nu)*(POW2(xi_nu)-1.));
				/* >>>chng 00 jul 27, avoid passing functions to macro so set to local temp var */
				xxx = 1. + 1./sqrt(4.*tau+3.*nu);
				eta1 = POW2(xxx)*exp(-theta_nu/tau);
				xi1 = (1. + nu/(2.*tau))*(1. + 1./sqrt(3./(2.*tau)+3.*nu))*exp(-theta_nu/tau);
			}
			sum += frac;
			*eta += frac*eta1;
			*xi += frac*xi1;
		}
		*eta /= sum;
		*xi /= sum;
	}

#	ifdef DEBUG_FUN
	fputs( " <->GrainScreen()\n", debug_fp );
#	endif
	return;
}


/* update items that depend on grain potential */
static void UpdatePot(long int nd)
{
	long i,
	  ion,
	  ipZ;
	float frac;

#	ifdef DEBUG_FUN
	fputs( "<+>UpdatePot()\n", debug_fp );
#	endif

	/* >>chng 00 jul 18, use finite range for bandgap to kick in, this aids
	 * in the proper convergence of the grain potential and heating, PvH */
	frac = ( gv.bin[nd]->dstpot > -DELTA_BANDGAP && gv.bin[nd]->dstpot < 0.f ) ?
		-gv.bin[nd]->dstpot/DELTA_BANDGAP : 1.f;

	gv.bin[nd]->ThresInf  = ( gv.bin[nd]->dstpot >= 0.f ) ? gv.bin[nd]->DustWorkFcn+gv.bin[nd]->dstpot :
		gv.bin[nd]->DustWorkFcn-frac*gv.bin[nd]->BandGap+gv.bin[nd]->TunnelCoef*gv.bin[nd]->dstpot;
	/* replace DustWorkFcn with ThresSurf where appropriate !!!! */
	gv.bin[nd]->ThresSurf = ( gv.bin[nd]->dstpot >= 0.f ) ? gv.bin[nd]->DustWorkFcn :
		gv.bin[nd]->DustWorkFcn-frac*gv.bin[nd]->BandGap+gv.bin[nd]->TunnelCoef*gv.bin[nd]->dstpot;
	gv.bin[nd]->PotSurf = ( gv.bin[nd]->dstpot >= 0.f ) ? gv.bin[nd]->dstpot :
		(1.f-gv.bin[nd]->TunnelCoef)*gv.bin[nd]->dstpot;
	/* PotInf is electron affinity in Weingartner & Draine
	 * gv.bin[nd]->PotInf = gv.bin[nd]->ThresInf + gv.bin[nd]->PotSurf; */
	/* >>chng 01 jan 08, ThresInf can become zero in initial stages of grain pot. search, PvH */
	gv.bin[nd]->ipDustPot = ipoint((double)MAX2(gv.bin[nd]->ThresInf,rfield.anu[0]));

	/* >>chng 00 jul 17, use description of Weingartner & Draine, 2000 */
	for( i=gv.bin[nd]->ipDustPot-1; i < rfield.nflux; i++ )
	{
		double Theta,x,y0,y2,Elo,Ehi;

		Theta = rfield.anu[i] - gv.bin[nd]->ThresSurf;
		x = MAX2(Theta/gv.bin[nd]->DustWorkFcn,0.);
		if( gv.bin[nd]->matType == MAT_CAR ) 
		{
			/* >>ref Bakes & Tielens, 1994, ApJ, 427, 822 */
			x = POW3(x)*POW2(x);
			y0 = 9.e-3*x/(1.+3.7e-2*x);
		}
		else if( gv.bin[nd]->matType == MAT_SIL ) 
		{
			/* >>ref Weingartner & Draine, 2000 */
			y0 = 0.5*x/(1.+5.*x);
		}
		else 
		{
			y0 = 0.;
		}
		if( gv.bin[nd]->dstpot > 0.f ) {
			Elo = -gv.bin[nd]->PotSurf;
			Ehi = MAX2(rfield.anu[i] - gv.bin[nd]->ThresInf,0.);
			y2 = POW2(Ehi) * (Ehi - 3.*Elo) / POW3(Ehi - Elo);
		}
		else 
		{
			y2 = 1.;
		}
		gv.bin[nd]->yhat[i] = (float)(y2*MIN2(y0*gv.bin[nd]->y1[i],1.));
	}

	/* this is charge per grain in elementary units */
	gv.bin[nd]->DustZ = (float)(gv.bin[nd]->dstpot*EVRYD*gv.bin[nd]->Capacity/ELEM_CHARGE);

	/* >>chng 00 jul 05, determine ionization stage Z0 the ion recombines to */
	for( ipZ=0; ipZ < LIMELM; ipZ++ )
	{
		if( abundances.lgElmtOn[ipZ])
		{
			ion = 0;
			while( ion <= ipZ && rfield.anu[Heavy.ipHeavy[ion][ipZ]-1] <= gv.bin[nd]->ThresSurf )
			{
				ion++;
			}
			gv.bin[nd]->RecomZ0[ipZ] = ion;
		}
		else
		{
			gv.bin[nd]->RecomZ0[ipZ] = LONG_MAX;
		}
	}
	
#	ifdef DEBUG_FUN
	fputs( " <->UpdatePot()\n", debug_fp );
#	endif
	return;
}

/* GrainTemperature computes grains temperature, and gas cooling */
static void GrainTemperature(long int nd,
			     /*@out@*/ float *dccool,
			     /*@out@*/ double *hcon,
			     /*@out@*/ double *hots,
			     /*@out@*/ double *hla,
			     /*@out@*/ double *dtherm,
			     /*@out@*/ double *thermionic)
{
	long int i,
	  ip,
	  ipLya;
	double CoolGrainThermionic,
	  eyhat,
	  norm,
	  rate,
	  x, 
	  y;
	float dcheat;

#	ifdef DEBUG_FUN
	fputs( "<+>GrainTemperature()\n", debug_fp );
#	endif

	/* direct heating by incident continuum (non-ionizing enegies only, dtherm contains the rest) */
	*hcon = 0.;
	/* heating by diffuse ots fields */
	*hots = 0.;
	/* heating by Ly alpha alone, for output only, is already included in hots */
	*hla = 0.;
	ipLya = HydroLines[0][ipH2p][ipH1s].ipCont;

	/* integrate over incident continuum for non-ionizing energies */
	for( i=0; i < gv.bin[nd]->ipDustPot-1; i++ )
	{
		*hcon += rfield.flux[i]*gv.bin[nd]->dstab1[i]*rfield.anu[i];
		*hots += rfield.SummedDif[i]*gv.bin[nd]->dstab1[i]*rfield.anu[i];

		/* >>chng 00 apr 18, moved calculation of hla, by PvH */
		if( i == ipLya-1 ) 
		{
			/*  heating by Ly A on dust in this zone,
			 *  only used for printout; Ly-a is already in OTS fields */
			*hla = rfield.otslin[i]*gv.bin[nd]->dstab1[i]*0.75;
		}

		/* fill in auxiliary array for quantum heating routine
		 * it reshuffles the input spectrum times the cross section to take
		 * the photo-electric effect into account. this prevents the quantum
		 * heating routine from having to calculate this effect over and over
		 * again; it can do a straight integration instead, making the code
		 * a lot simpler and faster. this initializes the array for non-ionizing
		 * energies, the reshuffling for higher energies is done in the next loop
		 * phiTilde has units events/H/s/cell at default depletion */
		if( gv.bin[nd]->lgQHeat ) 
		{
			gv.bin[nd]->phiTilde[i] = rfield.SummedCon[i]*gv.bin[nd]->dstab1[i];
		}
	}

	/* integrate over ionizing continuum; energy goes to dust and gas
	 * GasHeatPhotoEl is what heats the gas */
	gv.bin[nd]->GasHeatPhotoEl = 0.;

	/*  dtherm is part of h.nu-IP which heats dust before ejection */
	*dtherm = 0.;

	for( i=gv.bin[nd]->ipDustPot-1; i < rfield.nflux; i++ )
	{
		double DirectHeating,ehat,Elo,Ehi;

		/* >>chng 00 jul 17, use description of Weingartner & Draine, 2000 */
		if( gv.bin[nd]->dstpot >= 0.f ) 
		{
			Elo = -gv.bin[nd]->PotSurf;
			Ehi = MAX2(rfield.anu[i] - gv.bin[nd]->ThresInf,0.);
			ehat = Ehi/2. * (Ehi-2.*Elo)/(Ehi-3.*Elo);
		}
		else 
		{
			Elo = -gv.bin[nd]->PotSurf;
			Ehi = rfield.anu[i] - gv.bin[nd]->ThresInf - gv.bin[nd]->PotSurf;
			ehat = (Elo + Ehi)/2.;
		}
		eyhat = ehat*gv.bin[nd]->yhat[i];
		/*  DirectHeating is part of h.nu-IP which heats dust before ejection */
		DirectHeating = rfield.anu[i] - gv.bin[nd]->ThresSurf*gv.bin[nd]->yhat[i] - eyhat;

		/* this is heating by incident radiation field */
		*dtherm += rfield.flux[i]*gv.bin[nd]->dstab1[i]*DirectHeating;

		/*  this is heating by all diffuse fields:
		 *  SummedDif has all continua and lines */
		*hots += rfield.SummedDif[i]*gv.bin[nd]->dstab1[i]*DirectHeating;

		/*  GasHeatPhotoEl is rate grain photoionization heats the gas */
		gv.bin[nd]->GasHeatPhotoEl += rfield.SummedCon[i]*gv.bin[nd]->dstab1[i]*eyhat;

		/* >>chng 00 apr 18, include photo-electric effect, by PvH */
		if( i == ipLya-1 ) 
		{
			/*  heating by Ly A on dust in this zone,
			 *  only used for printout; Ly-a is already in OTS fields */
			*hla = rfield.otslin[i]*gv.bin[nd]->dstab1[i]*DirectHeating;
		}

		/* correct treatment of collisional processes in quantum heating */
		fixit(); /* factor in grain potential in eyhat and GasHeatPhotoEl !!! */
		if( gv.bin[nd]->lgQHeat ) 
		{
			/* this accounts for the photons that are fully absorbed by grain */
			gv.bin[nd]->phiTilde[i] = rfield.SummedCon[i]*gv.bin[nd]->dstab1[i]*
				(1.-gv.bin[nd]->yhat[i]);
			x = MAX2(DirectHeating,(double)rfield.anu[0]);
			ip = ipoint(x);
			/* this accounts for photons that eject an electron */
			gv.bin[nd]->phiTilde[ip-1] += rfield.SummedCon[i]*gv.bin[nd]->dstab1[i]*
				gv.bin[nd]->yhat[i];
		}
	}

	/* norm is used to convert all heating rates to erg/cm^3/s */
	norm = EN1RYD*phycon.hden*gv.bin[nd]->dstAbund;

	/* hcon is radiative heating by NON-ionizing radiation */
	*hcon *= norm;

	/* hots is total heating of the grain by diffuse fields */
	*hots *= norm;

	/* dtherm is part of h.nu-IP which heats dust before ejection */
	*dtherm *= norm;

	/* heating by Ly alpha alone, for output only, is already included in hots */
	*hla *= norm;

	/* this is total photon heating represented in phiTilde */
	gv.bin[nd]->GrainHeatPhi = *hcon + *hots + *dtherm;

	/* GasHeatPhotoEl is what heats the gas */
	if( gv.lgDHetOn )
	{
		gv.bin[nd]->GasHeatPhotoEl *= norm;
	}
	else
	{
		gv.bin[nd]->GasHeatPhotoEl = 0.;
	}

	/* now add in thermionic emissions (thermal evaporation of electrons)
	 * it gives a cooling term for the grain and the emission rate should
	 * be subtracted from CoolPotential since it counteracts electron collisions */
	rate = gv.bin[nd]->ThermRate*gv.bin[nd]->IntArea/4.*phycon.hden*gv.bin[nd]->dstAbund;
	CoolGrainThermionic = rate * (2.*BOLTZMANN*gv.bin[nd]->tedust + gv.bin[nd]->ThresSurf*EN1RYD);

	/* kinetic energy of electron is 2kT at surface of grain or at infinity */
	fixit(); /* is there no PotSurf term (or at least dependence) here ?? */

	*thermionic = rate * 2.*BOLTZMANN*gv.bin[nd]->tedust;

	/*  heating by thermal collisions with gas does work
	 *  DCHEAT is grain collisional heating by gas
	 *  DCCOOL is gas cooling due to collisions with grains
	 *  they are different since grain surface recombinations
	 *  heat the grains, but do not cool the gas ! */
	GrainCollHeating(nd,&dcheat,dccool);

	/*  dust often hotter than gas during initial TE search */
	if( nzone <= 2 )
		dcheat = MAX2(0.f,dcheat);

	/*  find power absorbed by dust and resulting temperature
	 *
	 * hcon is heating from incident continuum by non-ionizing radiation
	 * hots is heating from ots continua and lines
	 * dcheat is grain collisional heating by gas
	 * dtherm is part of h.nu-IP which heats dust before ejection,
	 * this is only direct heating by ionizing radiation
	 *
	 * GrainHeat is total heating of this grain type, to be balanced by cooling
	 * GrainHeat is the grain heat input - this is the actual number entering
	 * the energy balance */
	gv.bin[nd]->GrainHeat = *hcon + *hots + dcheat + *dtherm - CoolGrainThermionic;

	/* remember collisional heating for this grain species */
	gv.bin[nd]->GrainHeatColl = dcheat;

	/*  now find temperature, GrainHeat is sum of total heating of grain
	 *  >>chng 97 jul 17, divide by abundance here */
	x = log(MAX2(DBL_MIN,gv.bin[nd]->GrainHeat/phycon.hden/(double)gv.bin[nd]->dstAbund));

	/* >>chng 96 apr 27, as per Peter van Hoof comment */
	if( x < gv.bin[nd]->dstems[0] )
	{
		gv.bin[nd]->tedust = (float)GRAIN_TMIN;
	}
	else if( x > gv.bin[nd]->dstems[NDEMS-1] )
	{
		gv.bin[nd]->tedust = (float)GRAIN_TMAX;
	}
	else
	{
		splint(gv.bin[nd]->dstems,gv.dsttmp,gv.bin[nd]->dstslp,NDEMS,x,&y);
		gv.bin[nd]->tedust = (float)exp(y);
	}

	/*  save for later possible printout */
	gv.bin[nd]->TeGrainMax = (float)MAX2(gv.bin[nd]->TeGrainMax,gv.bin[nd]->tedust);

	if( trace.lgTrace && trace.lgDustBug )
	{
		fprintf( ioQQQ, " GrainTemperature finds %s Tdst%10.2e totHeat%10.2e Cont%10.2e ",
			 gv.bin[nd]->chDstLab, gv.bin[nd]->tedust, gv.bin[nd]->GrainHeat,
			 *hcon/gv.bin[nd]->GrainHeat);
		fprintf( ioQQQ, "OTS%10.2e coll%10.2e pho ej%10.2e\n", 
			 *hots/gv.bin[nd]->GrainHeat, dcheat/gv.bin[nd]->GrainHeat,
			 *dtherm/gv.bin[nd]->GrainHeat );
	}

#	ifdef DEBUG_FUN
	fputs( " <->GrainTemperature()\n", debug_fp );
#	endif
	return;
}

/* GrainCollHeating compute grains collisional heating cooling */
static void GrainCollHeating(long int nd,
			     /*@out@*/ float *dcheat, 
			     /*@out@*/ float *dccool)
{
	long int i,
	  ion, 
	  ipZ;
	double Accommodation,
	  atoms,
	  CollisionRateElectr,      /* rate electrons strike grains */
	  CollisionRateMol,         /* rate molecules strike grains */
	  CollisionRateIon,         /* rate ions strike grains */
	  cool,
	  CoolBounce,
	  CoolEmitted,
	  CoolElectrons,
	  CoolMolecules,
	  CoolPotential,
	  CoolPotential2,
	  CoolPotentialGas,
	  eta, 
	  heat,
	  HeatBounce,
	  HeatCollisions,
	  HeatElectrons,
	  HeatMolecules,
	  HeatRecombination, /* sum of abundances of ions times velocity times ionization potential times eta */
	  p_rad,
	  Stick,
	  WeightElec,
	  WeightMol,
	  xi;


#	ifdef DEBUG_FUN
	fputs( "<+>GrainCollHeating()\n", debug_fp );
#	endif


	/* this subroutine evaluates the gas heating-cooling rate
	 * (erg cm^-3 s^-1) due to grain gas collisions.
	 * the net effect can be positive or negative,
	 * depending on whether the grains or gas are hotter
	 * the physics is described in Baldwin, Ferland, Martin et al., 1991, ApJ 374, 580 */

	/* ============================================================================= */
	/* heating/cooling due to neutrals and positive ions */

	/* following loop sums up collisions between these nuclei
	 * electrons are done later
	 * HEAT will be rate collisions heat the grain
	 * COOL will be rate collisions cool the gas kinetics */
	heat = 0.;
	cool = 0.;
	/* loop over all stages of ionization */
	for( ion=0; ion <= LIMELM; ion++ )
	{
		/* this is heating of grains due to recombination energy of species,
		 * and assumes that every ion is fully neutralized upon striking the grain surface.
		 * all radiation produced in the recombination process is absorbed within the grain
		 *
		 * ion=0 are neutrals, ion=1 are single ions, etc
		 * each population is weighted by the AVERAGE velocity
		 * */
		CollisionRateIon = 0.;
		CoolPotential = 0.;
		CoolPotentialGas = 0.;
		HeatRecombination = 0.;

		/* >>chng 00 jul 19, replace classical results with results including image potential
		 * to correct for polarization of the grain as charged particle approaches. */
		GrainScreen(nd,ion,&eta,&xi);

		for( ipZ=MAX2(0,ion-1); ipZ < LIMELM; ipZ++ )
		{
			double CollisionRateOne;

			/* >>chng 00 apr 05, use correct accomodation coefficient, by PvH
			 * the coefficient is defined at the end of appendix A.10 of BFM
			 * assume ion sticking prob is unity */
			Stick = ( ion == gv.bin[nd]->RecomZ0[ipZ] ) ? gv.bin[nd]->AccomCoef[ipZ] : STICK_ION;
			/* this is rate with which charged ion strikes grain */
			/* >>chng 00 may 02, this had left 2./SQRTPI off */
			/* >>chng 00 may 05, use average speed instead of 2./SQRTPI*Doppler, by PvH */
			CollisionRateOne = Stick*xIonFracs[ion+1][ipZ]*gv.AveVel[ipZ];
			CollisionRateIon += CollisionRateOne;
			/* >>chng 00 jul 05, ion may not recombine to neutral, PvH */
			if( ion != gv.bin[nd]->RecomZ0[ipZ] ) 
			{
				/* grain surface recombination rates, used in MakeRecomb
				 * units s^-1 since has grain area in product, eta depends on ion */
				/* >>chng 00 jul 05, there should be a copy of this for each grain type, PvH */
				gv.bin[nd]->GrainRecom[ion][ipZ] = (float)(Stick*gv.AveVel[ipZ]*eta*
					gv.bin[nd]->IntArea/4.*phycon.hden*gv.bin[nd]->dstAbund);
			}
			else 
			{
				gv.bin[nd]->GrainRecom[ion][ipZ] = 0.;
			}
			CoolPotential += CollisionRateOne * (double)ion;
			CoolPotentialGas += CollisionRateOne * (double)gv.bin[nd]->RecomZ0[ipZ];
			/* this is sum of all energy liberated as ion recombines to Z0 charge inside grain */
			/* >>chng 00 jul 05, subtract energy needed to get electron out of grain potential, PvH */
			if( abundances.lgElmtOn[ipZ])
			{		
				if( ion > gv.bin[nd]->RecomZ0[ipZ] ) 
				{
					for( i=gv.bin[nd]->RecomZ0[ipZ]; i < ion; ++i ) 
					{
						HeatRecombination += CollisionRateOne*
							(rfield.anu[Heavy.ipHeavy[i][ipZ]-1]-gv.bin[nd]->ThresSurf);
					}
				}
				else if( ion < gv.bin[nd]->RecomZ0[ipZ] ) 
				{
					for( i=ion; i < gv.bin[nd]->RecomZ0[ipZ]; ++i ) 
					{
						HeatRecombination += CollisionRateOne*
							(gv.bin[nd]->ThresSurf-rfield.anu[Heavy.ipHeavy[i][ipZ]-1]);
					}
				}
			}
		}

		/* >>chng 00 may 01, boltzmann factor had multiplied all of factor instead
		 * of only first and last term.  pvh */

		/* equation 29 from Balwin et al 91 */
		/* this is direct collision rate, 2kT * xi, first term in eq 29 */
		HeatCollisions = CollisionRateIon * 2.*BOLTZMANN*phycon.te*xi;
		/* this is change in energy due to charge acceleration within grain's potential 
		 * this is exactly balanced by deceleration of incoming electrons and accelaration
		 * of outgoing photo-electrons and thermionic emissions; all these terms should
		 * add up to zero (total charge of grain should remain constant) */
		CoolPotential *= gv.bin[nd]->PotSurf*eta*EN1RYD;
		CoolPotentialGas *= gv.bin[nd]->PotSurf*eta*EN1RYD;
		/* this is recombination energy released within grain */
		HeatRecombination *= eta*EN1RYD;
		/* energy carried away by neutrals after recombination, so a cooling term */
		CoolEmitted = CollisionRateIon * 2.*BOLTZMANN*gv.bin[nd]->tedust*eta;

		/* total GraC 0 in the emission line output */
		heat += HeatCollisions - CoolPotential + HeatRecombination - CoolEmitted;

		/* rate kinetic energy lost from gas - gas cooling - eq 32 in BFM */
		/* this GrGC 0 in the main output */
		/* >>chng 00 may 05, reversed sign of gas cooling contribution */
		cool += HeatCollisions - CoolEmitted - CoolPotentialGas;
	}

	if( trace.lgTrace && trace.lgDustBug )
	{
		fprintf( ioQQQ, " ions heat/cool: %.4e %.4e\n", heat, cool );
	}

	/* ============================================================================= */
	/* heating/cooling due to electrons */

	ion = -1;
	/* >>chng 00 jul 19, replace classical results with results including image potential
	 * to correct for polarization of the grain as charged particle approaches. */
	GrainScreen(nd,ion,&eta,&xi);

	/* electron arrival rate - eqn 29 again */
	/*VelElectron = phycon.sqrte*6.2124e5*phycon.eden;*/
	/* sqrt term is average speed of electron */
	CollisionRateElectr = phycon.eden*sqrt(8.*BOLTZMANN/PI/ELECTRON_MASS*phycon.te);
	/* >>chng 00 jul 13, new sticking probability for electrons */
	if( gv.bin[nd]->dstpot <= 0. ) 
	{
		atoms = gv.bin[nd]->AvVol*gv.bin[nd]->dustp[0]/ATOMIC_MASS_UNIT/gv.bin[nd]->atomWeight;
		p_rad = 1./(1.+exp(20.-atoms));
	}
	else 
	{
		p_rad = 1.;
	}
	/* the second term is chance that electron passes through grain,
	 * 1-p_rad is chance that electron is ejected before grain settles
	 * see discussion in Weingartner & Draine, 2000, ApJ ....  */
	Stick = STICK_ELEC*(1. - exp(-gv.bin[nd]->AvRadius/MEAN_PATH))*p_rad;
	HeatCollisions = Stick*CollisionRateElectr*2.*BOLTZMANN*phycon.te*xi;

	/* this is change in energy due to charge acceleration within grain's potential 
	 * this term (perhaps) adds up to zero when summed over all charged particles */
	CoolPotential = Stick*CollisionRateElectr * (double)ion*gv.bin[nd]->PotSurf*eta*EN1RYD;

	/* >>chng 00 jul 05, this is term for energy released due to recombination, PvH */
	HeatRecombination = Stick*CollisionRateElectr * gv.bin[nd]->ThresSurf*eta*EN1RYD;

	/* add heating due to recombination of electrons in phiTilde */
	if( gv.bin[nd]->lgQHeat ) 
	{
		double Einf,Ehi,Ek,fac,rate,ylo,yhi;

		fac = BOLTZMANN/EN1RYD*phycon.te;
		/* this is rate in events/H/s at standard depletion */
		rate = Stick*CollisionRateElectr*gv.bin[nd]->IntArea/4.*eta;
		Einf = gv.bin[nd]->DustWorkFcn+gv.bin[nd]->PotSurf;

		Ek = gv.bin[nd]->ThresInf - Einf;
		yhi = -(Ek/fac+1.)*exp(-Ek/fac);
		for( i=0; i < rfield.nflux; i++ ) 
		{
			ylo = yhi;
			Ehi = rfield.anu[i] + 0.5*rfield.widflx[i];
			if( Ehi >= gv.bin[nd]->ThresInf ) 
			{
				/* this is kinetic energy of electron at infinity */
				Ek = Ehi - Einf;
				yhi = -(Ek/fac+1.)*exp(-Ek/fac);
				gv.bin[nd]->phiTilde[i] += rate*(yhi-ylo);
			}
		}
		/* collisional processes in quantum heating not correctly treated */
		fixit(); /* total amount of heating not correct for negative grains due to ylo not being -1 ???? */
		gv.bin[nd]->GrainHeatPhi += (HeatCollisions - CoolPotential + HeatRecombination)*
			gv.bin[nd]->IntArea/4.*phycon.hden*gv.bin[nd]->dstAbund;
	}
			
	/* >>chng 00 jul 05, add in terms for electrons that bounce off grain, PvH */
	WeightElec = ELECTRON_MASS/ATOMIC_MASS_UNIT;
	Accommodation = 2.*gv.bin[nd]->atomWeight*WeightElec/POW2(gv.bin[nd]->atomWeight+WeightElec);
	HeatBounce = (1. - Stick)*Accommodation*CollisionRateElectr * 2.*BOLTZMANN*phycon.te*xi;
	CoolBounce = (1. - Stick)*Accommodation*CollisionRateElectr * 2.*BOLTZMANN*gv.bin[nd]->tedust*eta;
	CoolPotential2 = (1. - Stick)*Accommodation*CollisionRateElectr * (double)ion*gv.bin[nd]->PotSurf*eta*EN1RYD;

	/* >>chng 00 may 02, CoolPotential had not been included */
	/* >>chng 00 jul 05, HeatRecombination had not been included */
	HeatElectrons = HeatCollisions-CoolPotential+HeatRecombination+HeatBounce-CoolPotential2-CoolBounce;
	heat += HeatElectrons;

	CoolElectrons = HeatCollisions+HeatBounce-CoolBounce-CoolPotential2;
	cool += CoolElectrons;

	if( trace.lgTrace && trace.lgDustBug )
	{
		fprintf( ioQQQ, " electrons heat/cool: %.4e %.4e\n", HeatElectrons, CoolElectrons );
	}

	/* ============================================================================= */
	/* heating/cooling due to molecules */

	/* sticking prob for H2 onto grain,
	 * estimated from accomodation coefficient defined at end of A.10 in BFM */
	WeightMol = 2.*AtomcWgt.AtomicWeight[0];
	Accommodation = 2.*gv.bin[nd]->atomWeight*WeightMol/POW2(gv.bin[nd]->atomWeight+WeightMol);
	/* molecular hydrogen onto grains */
	CollisionRateMol = Accommodation*hmi.htwo*
		sqrt(8.*BOLTZMANN/PI/ATOMIC_MASS_UNIT/WeightMol*phycon.te);

	/* now add in CO */
	WeightMol = AtomcWgt.AtomicWeight[5] + AtomcWgt.AtomicWeight[7];
	Accommodation = 2.*gv.bin[nd]->atomWeight*WeightMol/POW2(gv.bin[nd]->atomWeight+WeightMol);
	CollisionRateMol += Accommodation*hevmolec.hevmol[IPCO-1]*
		sqrt(8.*BOLTZMANN/PI/ATOMIC_MASS_UNIT/WeightMol*phycon.te);

	/* xi and eta are unity for neutrals and so ignored */
	HeatCollisions = CollisionRateMol * 2.*BOLTZMANN*phycon.te;
	CoolEmitted = CollisionRateMol * 2.*BOLTZMANN*gv.bin[nd]->tedust;

	HeatMolecules = HeatCollisions - CoolEmitted;
	heat += HeatMolecules;

	/* >>chng 00 may 05, reversed sign of gas cooling contribution */
	CoolMolecules = HeatCollisions - CoolEmitted;
	cool += CoolMolecules;

	if( trace.lgTrace && trace.lgDustBug )
	{
		fprintf( ioQQQ, " molecules heat/cool: %.4e %.4e\n", HeatMolecules, CoolMolecules );
	}

	/* ============================================================================= */

	/* heating/cooling above is in erg/s/cm^2 -> multiply with projected grain area per cm^3 */
	/* GraC 0 is integral of dcheat, the total collisional heating of the grain */
	*dcheat = (float)(heat*gv.bin[nd]->IntArea/4.*phycon.hden*gv.bin[nd]->dstAbund);

	/* GrGC 0 total cooling of gas integrated */
	*dccool = (float)(cool*gv.bin[nd]->IntArea/4.*phycon.hden*gv.bin[nd]->dstAbund);

#	ifdef DEBUG_FUN
	fputs( " <->GrainCollHeating()\n", debug_fp );
#	endif
	return;
}

/*GrainDrift compute grains drift velocity */
static void GrainDrift(void)
{
	long int i, 
	  loop, 
	  nd;

	double alam, 
	  corr, 
	  dmomen, 
	  fac, 
	  fdrag, 
	  g0, 
	  g2, 
	  phi2lm, 
	  psi, 
	  rdust, 
	  si, 
	  vdold, 
	  volmom;

#	ifdef DEBUG_FUN
	fputs( "<+>GrainDrift()\n", debug_fp );
#	endif

	for( nd=0; nd < gv.nBin; nd++ )
	{
		/* find momentum absorbed by grain */
		dmomen = 0.;
		for( i=0; i < rfield.nflux; i++ )
		{
			dmomen += (rfield.flux[i]+rfield.ConOutInter[i]+rfield.ConOutRecInter[i]+rfield.outlin[i])*
			  rfield.anu[i]*(gv.bin[nd]->dstab1[i] + gv.bin[nd]->dstsc1[i]);
		}
		dmomen *= EN1RYD*4./gv.bin[nd]->IntArea;

		/* now find force on grain, and drift velocity */
		fac = 2*BOLTZMANN*phycon.te;

		/* now PSI defined by Draine and Salpeter 79 Ap.J. 231, 77 (1979) */
		psi = gv.bin[nd]->dstpot*TE1RYD/phycon.te;
		if( psi > 0. )
		{
			rdust = 1.e-6;
			alam = log(20.702/rdust/psi*phycon.sqrte/phycon.SqrtEden);
		}
		else
		{
			alam = 0.;
		}

		phi2lm = POW2(psi)*alam;
		corr = 2.;
		for( loop = 0; loop < 10 && fabs(corr-1.) > 0.03; loop++ )
		{
			vdold = gv.bin[nd]->DustDftVel;

			/* interactions with protons */
			si = gv.bin[nd]->DustDftVel/phycon.sqrte*7.755e-5;
			g0 = 1.5045*si*sqrt(1.+0.4418*si*si);
			g2 = si/(1.329 + POW3(si));

			/* drag force due to protons, both linear and square in velocity
			 * equation 4 from D+S Ap.J. 231, p77. */
			fdrag = fac*h.hii*(g0 + phi2lm*g2);

			/* drag force due to interactions with electrons */
			si = gv.bin[nd]->DustDftVel/phycon.sqrte*1.816e-6;
			g0 = 1.5045*si*sqrt(1.+0.4418*si*si);
			g2 = si/(1.329 + POW3(si));
			fdrag += fac*phycon.eden*(g0 + phi2lm*g2);

			/* drag force due to collisions with hydrogen and helium atoms */
			si = gv.bin[nd]->DustDftVel/phycon.sqrte*7.755e-5;
			g0 = 1.5045*si*sqrt(1.+0.4418*si*si);
			fdrag += fac*(h.hi + 1.1*xIonFracs[1][ipHELIUM])*g0;

			/* drag force due to interactions with helium ions */
			si = gv.bin[nd]->DustDftVel/phycon.sqrte*1.551e-4;
			g0 = 1.5045*si*sqrt(1.+0.4418*si*si);
			g2 = si/(1.329 + POW3(si));
			fdrag += fac*xIonFracs[2][ipHELIUM]*(g0 + phi2lm*g2);

			/* this term does not work
			 *  2      HEIII*(G0+4.*PSI**2*(ALAM-0.693)*G2) )
			 * this is total momentum absorbed by dust per unit vol */
			volmom = dmomen/SPEEDLIGHT;

			if( fdrag > 0. )
			{
				corr = sqrt(volmom/fdrag);
				gv.bin[nd]->DustDftVel = (float)(vdold*corr);
			}
			else
			{
				corr = 1.;
				negdrg.lgNegGrnDrg = TRUE;
			}

			if( trace.lgTrace && trace.lgDustBug )
			{
				fprintf( ioQQQ, "     %2ld new drift velocity:%10.2e momentum absorbed:%10.2e\n", 
				  loop, gv.bin[nd]->DustDftVel, volmom );
			}
		}
	}


#	ifdef DEBUG_FUN
	fputs( " <->GrainDrift()\n", debug_fp );
#	endif
	return;
}


/****************************************************************************
 *
 * qheat: driver routine for non-equilibrium heating of grains
 *
 * This routine calls qheat1 (which does the actual non-equilibrium
 * calculations) and does the subsequent quality control.
 *
 * Written by Peter van Hoof (UK, CITA).
 *
 ****************************************************************************/

void qheat(double qtemp[],  /* qtemp[NQGRID] */
	   double qprob[],  /* qprob[NQGRID] */
	   /*@out@*/ long int *qnbin,
	   long int nd)
{
	int lgBigChange=TRUE,
	  lgDelta=TRUE,
	  lgFailed=TRUE,
	  lgOK=FALSE;
	long int i;
	double dPdlnT[NQGRID],
	  new_tmin,
	  x1;

#	ifdef DEBUG_FUN
	fputs( "<+>qheat()\n", debug_fp );
#	endif

	if( trace.lgTrace && trace.lgDustBug )
	{
		fprintf( ioQQQ, " >>>>qheat called grain %s\n", gv.bin[nd]->chDstLab );
	}

	assert( gv.bin[nd]->lgQHeat );

	for( i=0; i < LOOP_MAX && !lgOK; i++ ) 
	{
		qheat1(qtemp,qprob,dPdlnT,qnbin,nd,&new_tmin,&lgFailed,&lgDelta);

		x1 = MAX2(new_tmin,gv.bin[nd]->qtmin)/MIN2(new_tmin,gv.bin[nd]->qtmin);
		lgBigChange = (x1 > 2.0);

		gv.bin[nd]->qtmin = new_tmin;

		lgOK = !lgFailed && !lgDelta && !lgBigChange;

		if( lgDelta ) 
		{
			if( gv.bin[nd]->qpres > 1.01*QHEAT_HIGH_RES ) {
				gv.bin[nd]->qpres /= 3.;
				gv.bin[nd]->qpres = MAX2(gv.bin[nd]->qpres,QHEAT_HIGH_RES);
			}
			else 
			{
				break;
			}
		}

		if( !lgOK && trace.lgTrace && trace.lgDustBug ) 
		{
			fprintf( ioQQQ, " >>>>qheat trying another iteration, lgFailed %c lgBigChange %c\n\n",
				 TorF(lgFailed), TorF(lgBigChange) );
		}
	}

	gv.bin[nd]->lgUseQHeat = lgOK;
	gv.bin[nd]->lgEverQHeat = gv.bin[nd]->lgEverQHeat || gv.bin[nd]->lgUseQHeat;

	if( !lgOK ) 
	{
		*qnbin = 0;
	}

	if( gv.QHPunchFile != NULL && ( IterCnt.lgLastIt || !gv.lgQHPunLast ) ) 
	{
		fprintf( gv.QHPunchFile, "\nDust Temperature Distribution: grain %s zone %ld\n",
			 gv.bin[nd]->chDstLab,nzone );

		fprintf( gv.QHPunchFile, "Equilibrium temperature: %.2f\n", gv.bin[nd]->tedust );

		if( lgOK ) 
		{
			fprintf( gv.QHPunchFile, "Number of bins: %ld\n", *qnbin-1 );
			fprintf( gv.QHPunchFile, "  Tgrain     probability    dP/dlnT\n" );
			for( i=1; i < *qnbin; i++ ) 
			{
				fprintf( gv.QHPunchFile, "%-12.4e %-12.4e %-12.4e\n",
					 qtemp[i],qprob[i],dPdlnT[i] );
			}
		}
		else 
		{
			fprintf( gv.QHPunchFile, "**** quantum heating was not used\n" );
		}
	}

	if( trace.lgTrace && trace.lgDustBug )
	{
		if( lgOK ) 
		{
			fprintf( ioQQQ, " >>>>qheat converged\n\n");
		}
		else if( lgDelta ) 
		{
			fprintf( ioQQQ, " >>>>qheat found delta function, resolution %.4e\n\n",
				 gv.bin[nd]->qpres);
		}
		else 
		{
			fprintf( ioQQQ, " >>>>qheat failed, lgFailed %c lgBigChange %c\n",
				 TorF(lgFailed), TorF(lgBigChange) );
		}
	}

#	ifdef DEBUG_FUN
	fputs( " <->qheat()\n", debug_fp );
#	endif
	return;
}

/****************************************************************************
 *
 * qheat1: main routine for calculating non-equilibrium heating of grains
 *
 * This routine implements the algorithm outlined in:
 *   Guhathakurtha & Draine, 1989, ApJ 345, 230
 *
 * The original (fortran) version of the code was written by Kevin Volk.
 *
 * Heavily modified and adapted for new style grains by Peter van Hoof.
 *
 ****************************************************************************/

static void qheat1(/*@out@*/ double qtemp[],  /* qtemp[NQGRID]  */
		   /*@out@*/ double qprob[],  /* qprob[NQGRID]  */
		   /*@out@*/ double dPdlnT[], /* dPdlnT[NQGRID] */
		   /*@out@*/ long int *qnbin,
		   long int nd,
		   /*@out@*/ double *new_tmin,
		   /*@out@*/ int *lgFailed,
		   /*@out@*/ int *lgDelta)
{
	long int ind,
	  j,
	  k, 
	  nbin;
	double AveEnergyColl,
	  AveEnergyPhot,
	  BoltzExp,
	  BoltzFac,
	  bval_jk,
	  GCollRate,
	  delta_x,
	  delta_y,
	  delu[NQGRID],
	  dlnpdU,
	  fac = 0.,
	  frac,
	  Lambda[NQGRID],
	  p[NQGRID],
	  Phi[NCELL],
	  NumEvents,
	  phi,
	  PhotRate,
	  qtmin1, 
	  RadCooling,
	  scale, 
	  scale2, 
	  sum,
	  sum2,
	  u1[NQGRID],
	  uhi,
	  ulo,
	  umin,
	  val,
	  val2,
	  x,
	  xx,
	  y,
	  ylo;


#	ifdef DEBUG_FUN
	fputs( "<+>qheat1()\n", debug_fp );
#	endif

	if( trace.lgTrace && trace.lgDustBug )
	{
		fprintf( ioQQQ, "  qheat1 called grain %s\n", gv.bin[nd]->chDstLab );
	}

	/* scale2 is to convert grain cross sections in cm^2/H at standard depletion
	 * (as supplied in dstab1) to cross sections in cm^2/cm^3 at actual depletion */
	scale2 = phycon.hden*gv.bin[nd]->dstAbund;
	/* scale is to convert heating/cooling rates from Ryd/s/cm^3 to Ryd/s/grain, in cm^3/grain */
	scale = gv.bin[nd]->qhnrat/scale2;

	xx = 0.;
	for( j=0; j < rfield.nflux; j++ )
	{
		/* >>chng 97 jul 17, to summed continuum */
		/* >>chng 00 mar 30, to phiTilde, to keep track of photo-electric effect, by PvH */
		xx += gv.bin[nd]->phiTilde[j]*scale2;
		/* Phi has units events/cm^3/s at actual depletion */
		Phi[j] = xx;
	}

	/* thermionic emission not treated in quantum heating */
	fixit(); /* add thermionic emissions in thermal balance */

	/* the average energy of a colliding particle is 2kTe, NOT 1.5kTe */
	AveEnergyColl = 2.*BOLTZMANN*phycon.te;
	/* This is the number rate at which particles hit grains, in collisions/cm^3/s */
	GCollRate = (gv.bin[nd]->GrainHeat-gv.bin[nd]->GrainHeatPhi)/AveEnergyColl;

	/* PhotRate is the rate at which photons hit grains, in photons/cm^3/s */
	PhotRate = Phi[rfield.nflux-1];
	AveEnergyPhot = gv.bin[nd]->GrainHeatPhi/PhotRate;
	x = ufunct((double)gv.bin[nd]->tedust,nd)*EN1RYD;
	frac = gv.bin[nd]->GrainHeatPhi/gv.bin[nd]->GrainHeat;
	NumEvents = frac*x/AveEnergyPhot + (1.-frac)*x/AveEnergyColl;

	qtmin1 = ( gv.bin[nd]->qtmin < gv.bin[nd]->tedust ) ? gv.bin[nd]->qtmin : 0.7*gv.bin[nd]->tedust;
	qtmin1 = MAX2(qtmin1,GRAIN_TMIN);

	if( trace.lgTrace && trace.lgDustBug )
	{
		fprintf( ioQQQ, "    photon, coll heating: %.4e %.4e, photon, coll rate %.4e %.4e scale %.4e\n",
			 gv.bin[nd]->GrainHeatPhi,gv.bin[nd]->GrainHeat-gv.bin[nd]->GrainHeatPhi,
			 PhotRate*scale,GCollRate*scale,scale);
		fprintf( ioQQQ, "    av grain temperature %.4e av grain enthalpy %.4e electr temp %.4e\n",
			 gv.bin[nd]->tedust,x,phycon.te);
		fprintf( ioQQQ, "    got lowest grain temp %.4e used lowest: %.4e\n",gv.bin[nd]->qtmin,qtmin1);
		fprintf( ioQQQ, "    average number of multiple events: %.4e, resolution %.4f\n",
			 NumEvents,gv.bin[nd]->qpres );
	}

	/* if average number of multiple photon events is too high -> don't even try */
	if( NumEvents > MAX_EVENTS ) 
	{
		*qnbin = 0;
		*new_tmin = qtmin1;
		*lgFailed = TRUE;
		*lgDelta = TRUE;

		if( trace.lgTrace && trace.lgDustBug )
		{
			fprintf( ioQQQ, "  qheat1 returns, NumEvents exceeding MAX_EVENTS\n" );
		}

#		ifdef DEBUG_FUN
		fputs( " <->qheat1()\n", debug_fp );
#		endif
		return;
	}

	ulo = rfield.AnuOrg[0];
	uhi = rfield.AnuOrg[rfield.nflux-1];

	nbin = NQGRID;
	qtemp[0] = qtmin1;
	/* u1 holds enthalpy in Ryd/grain */
	u1[0] = ufunct(qtemp[0],nd);
	/* >>chng 00 mar 22, taken out factor 4, factored in hden and dstAbund
	 * interpolate in dstems array instead of integrating explicitly, by PvH */
	xx = log(qtemp[0]);
	splint(gv.dsttmp,gv.bin[nd]->dstems,gv.bin[nd]->dstslp2,NDEMS,xx,&y);
	/* Lambda holds the radiated energy for grains in this bin, in Ryd/s/grain */
	Lambda[0] = exp(y)*scale2*scale/EN1RYD;
	/* set up first step of integration */
	delu[0] = gv.bin[nd]->qpres*u1[0]/100.;
	p[0] = PROB_CUTOFF_LO;
	qprob[0] = p[0]*delu[0];
	dPdlnT[0] = 0.;
	RadCooling = qprob[0]*Lambda[0];
	delu[1] = delu[0];
	BoltzFac = phycon.te/TE1RYD;

	for( k=1; k < NQGRID; k++ )
	{
		u1[k] = u1[k-1] + delu[k];
		qtemp[k] = inv_ufunct(u1[k],nd);
		xx = log(qtemp[k]);
		splint(gv.dsttmp,gv.bin[nd]->dstems,gv.bin[nd]->dstslp2,NDEMS,xx,&y);
		Lambda[k] = exp(y)*scale2*scale/EN1RYD;

		sum = 0.;
		sum2 = 0.;
		for( j=0; j < k; j++ )
		{
			umin = u1[k] - delu[k]/2. - u1[j];
			ylo = umin/BoltzFac;
			BoltzExp = exp(-ylo);
			val2 = 0.;

			/* val is the value of Phi at umin */
			if( umin < ulo ) 
			{
				val = 0.;
			}
			else if( umin > uhi ) 
			{
				val = Phi[rfield.nflux-1];
			}
			else {
				/* ind is on C scale */
				ind = ipoint(umin)-1;
				frac = (umin-(double)rfield.anu[ind])/(double)rfield.widflx[ind] + 0.5;
				y = ( ind == 0 ) ? 0. : Phi[ind-1];
				val = y + frac*(Phi[ind] - y);
				if( k <= NSTARTUP ) 
				{
					y = ( ind == 0 ) ? 0. : gv.bin[nd]->phiTilde[ind-1]/rfield.widflx[ind-1];
					val2 = y + frac*(gv.bin[nd]->phiTilde[ind]/rfield.widflx[ind] - y);
				}
			}

			/* bval_jk is the total excitation rate from j to k and higher due to photon 
			 * absorptions and particle collisions, it already implements Eq. 2.17
			 * of Guhathakurtha & Draine, in events/grain/s */
			/* >>chng 00 mar 27, factored in hden (in Phi), by PvH */
			/* >>chng 00 mar 29, add in contribution due to particle collisions, by PvH */
			bval_jk = ((Phi[rfield.nflux-1]-val) + GCollRate*(1.+ylo)*BoltzExp)*scale;
			sum += qprob[j]*bval_jk;

			if( k <= NSTARTUP ) 
			{
				phi = (val2*scale2 + GCollRate/BoltzFac*ylo*BoltzExp)*scale;
				sum2 += qprob[j]*phi;
			}
		}

		/* >>chng 00 mar 27, multiplied with delu, by PvH */
 		/* qprob[k] = sum*delu[k]/(Lambda[k] - Lambda[0]); */
		/* >>chng 00 apr 5, taken out Lambda[0], improves convergence at low end dramatically!, by PvH */
		p[k] = sum/Lambda[k];
		qprob[k] = p[k]*delu[k];
		dPdlnT[k] = qprob[k]/log(qtemp[k]/qtemp[k-1]);

		/* calculate logarithmic derivative to estimate new stepsize */
		if( k+1 < NQGRID ) 
		{
			/* this expression is more stable for the first few points */
			if( k <= NSTARTUP ) 
			{
				y = Phi[rfield.nflux-1]*scale;
				y = y - (Lambda[k] - Lambda[k-1])/delu[k];
				dlnpdU = (y - sum2/p[k])/Lambda[k];
			}
			else 
			{
				if( p[k] > p[k-1] ) 
				{
					dlnpdU = log(p[k]/p[k-1])/delu[k-1];
				}
				else 
				{
					dlnpdU = log(p[k]/p[k-1])/delu[k];
				}
			}
			xx = fabs(dlnpdU);
			/* the second expression limits the stepsize near the maximum */
			delu[k+1] = ( xx > 4./u1[k] ) ? gv.bin[nd]->qpres/xx : gv.bin[nd]->qpres*u1[k]/4.;
		}

		/* if qtmin is far too low, qprob can become extremely large, exceeding
		 * even double precision range. the following check should prevent overflows */
		if( qprob[k] > sqrt(DBL_MAX) ) 
		{
			for( j=0; j <= k; j++ ) 
			{
				p[j] /= DBL_MAX;
				qprob[j] /= DBL_MAX;
				dPdlnT[j] /= DBL_MAX;
			}
			RadCooling /= DBL_MAX;
		}
		RadCooling += qprob[k]*Lambda[k];

		/* force thermal equilibrium of the grains */
		fac = RadCooling/gv.bin[nd]->GrainHeat*EN1RYD/scale;

		/* this is regular stop criterion */
		if( dPdlnT[k] < dPdlnT[k-1] && dPdlnT[k]/fac < PROB_CUTOFF_HI ) 
		{
			nbin = k+1;
			break;
		}
	}

	if( nbin <= NQMIN ) 
	{
		*qnbin = 0;
		*new_tmin = qtmin1;
		*lgFailed = TRUE;
		*lgDelta = TRUE;

		if( trace.lgTrace && trace.lgDustBug )
		{
			fprintf( ioQQQ, "  qheat1 returns, nbin=%ld <= NQMIN\n", nbin );
		}

#		ifdef DEBUG_FUN
		fputs( " <->qheat1()\n", debug_fp );
#		endif
		return;
	}

	*qnbin = nbin;

	sum = 0.;
	for( j=0; j < nbin; j++ )
	{
		p[j] /= fac;
		qprob[j] /= fac;
		dPdlnT[j] /= fac;
		sum += qprob[j];
	}

	*lgFailed = ( dPdlnT[nbin-1]/sum > SAFETY*PROB_CUTOFF_HI );

	*new_tmin = 0.;
	for( j=1; j < nbin; j++ ) 
	{
		if( dPdlnT[j]/sum < PROB_CUTOFF_LO ) 
		{
			*new_tmin = qtemp[j];
		}
		else 
		{
			if( j == 1 ) 
			{
				*lgFailed = *lgFailed || ( dPdlnT[j]/sum > SAFETY*PROB_CUTOFF_LO );
				if( dPdlnT[1] < 0.999*dPdlnT[NSTARTUP] ) 
				{
					delta_x = log(qtemp[NSTARTUP]/qtemp[1]);
					delta_y = log(dPdlnT[NSTARTUP]/dPdlnT[1]);
					delta_x *= log(PROB_CUTOFF_LO*sum/dPdlnT[1])/delta_y;
					*new_tmin = exp(log(qtemp[1]) + delta_x);
				}
			}
			break;
		}
	}
	*new_tmin = MAX3(*new_tmin,qtmin1/DEF_FAC,GRAIN_TMIN);

	/* the fact that the probability normalization fails probably indicates
	 * that the distribution is too close to a delta function to resolve */
	*lgDelta = *lgFailed ? FALSE : (fabs(sum-1.) > PROB_TOL);
		
	if( trace.lgTrace && trace.lgDustBug )
	{
		fprintf( ioQQQ, "    nbin %ld total probability %.4e new_tmin %.4e lgFailed %c lgDelta %c\n",
			 nbin,sum,*new_tmin,TorF(*lgFailed),TorF(*lgDelta) );
		fprintf( ioQQQ, "  qheat1 returns\n\n" );
	}

#	ifdef DEBUG_FUN
	fputs( " <->qheat1()\n", debug_fp );
#	endif
	return;
}


/* calculate the enthalpy of a grain at a given temperature, in Ryd */
static double ufunct(double temp, 
		     long int nd)
{
	long int itype,
	  j;
	double atoms,
	  enthalpy;

#	ifdef DEBUG_FUN
	fputs( "<+>ufunct()\n", debug_fp );
#	endif

	if( temp <= 0. ) 
	{
		fprintf( ioQQQ, " ufunct called with non-positive temperature: %.6e\n" , temp );
		puts( "[Stop in ufunct]" );
		cdEXIT(1);
	}
	assert( nd >= 0 && nd < gv.nBin );

	itype = gv.bin[nd]->matType;
	atoms = gv.bin[nd]->AvVol*gv.bin[nd]->dustp[0]/ATOMIC_MASS_UNIT/gv.bin[nd]->atomWeight;

	enthalpy = 0.;
	if( itype == MAT_CAR )
	{
		/* enthalpy for pah/graphitic grains in Ryd
		 * derived from: Guhathakurta & Draine, 1989, ApJ 345, 230 */
		enthalpy = (4.15e-22/EN1RYD)*pow(temp,3.3)/(1. + 6.51e-03*temp + 
		  1.5e-06*temp*temp + 8.3e-07*pow(temp,2.3));
	}
	else if( itype == MAT_SIL )
	{
		/* enthalpy for silicate grains (and grey grains) in Ryd */
		/* >>chgn 00 Mar 22, changed run of j, highest bin was missed, by PvH */
		/* limits of tlim set above, 0 and DBL_MAX, so always OK */
		for( j = 0; j < 4 && temp > tlim[j]; j++ )
		{
			enthalpy += cval[j]*(pow(MIN2(temp,tlim[j+1]),power[j]) - pow(tlim[j],power[j]));
		}
	}
	else 
	{
		fprintf( ioQQQ, " ufunct called with unknown type for enthalpy function: %ld\n" , itype );
		puts( "[Stop in ufunct]" );
		cdEXIT(1);
	}

	/* >>chng 00 mar 23, use formula 3.1 of Guhathakurtha & Draine, by PvH */
	enthalpy *= atoms-2.;

	if( enthalpy <= 0. ) 
	{
		fprintf( ioQQQ, " ufunct finds non-positive enthalpy: %.6e, what's up?\n" , enthalpy );
		puts( "[Stop in ufunct]" );
		cdEXIT(1);
	}

#	ifdef DEBUG_FUN
	fputs( " <->ufunct()\n", debug_fp );
#	endif
	return( enthalpy );
}

/* this is the inverse of ufunct: determine grain temperature as a function of enthalpy */
static double inv_ufunct(double enthalpy, 
			 long int nd)
{
	long int itype;
	double atoms,
	  temp;

#	ifdef DEBUG_FUN
	fputs( "<+>inv_ufunct()\n", debug_fp );
#	endif

	if( enthalpy <= 0. ) 
	{
		fprintf( ioQQQ, " inv_ufunct called with non-positive enthalpy: %.6e\n" , enthalpy );
		puts( "[Stop in ufunct]" );
		cdEXIT(1);
	}
	assert( nd >= 0 && nd < gv.nBin );

	itype = gv.bin[nd]->matType;
	atoms = gv.bin[nd]->AvVol*gv.bin[nd]->dustp[0]/ATOMIC_MASS_UNIT/gv.bin[nd]->atomWeight;

	enthalpy /= atoms-2.;
	temp = 0.;
	if( itype == MAT_CAR )
	{
		double x0,x1,x2,denom,toler=1.e-6;

		/* inverse enthalpy for pah/graphitic grains in Ryd
		 * derived from: Guhathakurta & Draine, 1989, ApJ 345, 230 */
		x1 = pow(enthalpy*(EN1RYD/4.15e-22),1./3.3);
		x2 = pow(enthalpy*(EN1RYD/2.74e-19),1./1.9);
		x1 = MAX2(x1,x2);

		do 
		{
			x0 = x1;
			denom = 1. + 6.51e-03*x0 + 1.5e-06*x0*x0 + 8.3e-07*pow(x0,2.3);
			x1 = pow(enthalpy*denom*(EN1RYD/4.15e-22),1./3.3);
		} 
		while( fabs(x1-x0)/x1 > toler );

		temp = x1;
	}
	else if( itype == MAT_SIL )
	{
		double ulow;
		long int j;

		/* inverse enthalpy for silicate grains (and grey grains) in Ryd */
		for( j = 0; j < 4; j++ )
		{
			ulow = cval[j]*(pow(tlim[j+1],power[j]) - pow(tlim[j],power[j]));
			if( ulow < enthalpy ) 
			{
				enthalpy -= ulow;
			}
			else 
			{
				break;
			}
		}
		enthalpy += cval[j]*pow(tlim[j],power[j]);
		temp = pow(enthalpy/cval[j],1./power[j]);
	}
	else 
	{
		fprintf( ioQQQ, " inv_ufunct called with unknown type for enthalpy function: %ld\n" , itype );
		puts( "[Stop in inv_ufunct]" );
		cdEXIT(1);
	}

	if( temp <= 0. ) 
	{
		fprintf( ioQQQ, " inv_ufunct finds non-positive temperature: %.6e, what's up?\n" , temp );
		puts( "[Stop in inv_ufunct]" );
		cdEXIT(1);
	}

#	ifdef DEBUG_FUN
	fputs( " <->inv_ufunct()\n", debug_fp );
#	endif
	return( temp );
}
