/*OpacityAdd1Element enter total photo cross section for all subshells into opacity array */
#include "cddefines.h"
#include "hydrogenic.h"
#include "nh.h"
#include "rfield.h"
#include "nsshells.h"
#include "ionfracs.h"
#include "hydrophocs.h"
#include "fluors.h"
#include "hphoto.h"
#include "opacity.h"
/*opinPL add opacity of individual species, with stimulated emission, for power law opacity */
static void opinPL(
	long int n,	/* quantum number of level on C not physical scale, 
					 * must be less than NHYDRO_MAX_LEVEL, dim of STH */
  long int ipZ,/* charge, ipZ=0 for h, 1 for he, etc */
  float abundance, 
  double b);

void OpacityAdd1Element(
		/* ipZ is 0 for H, 1 for He, etc */
		long int ipZ)
{
	long int ipHi, 
	  ipop, 
	  low, 
	  n, 
	  nion, 
	  nshell;
	char chStat;
	double abundance;

#	ifdef DEBUG_FUN
	fputs( "<+>OpacityAdd1Element()\n", debug_fp );
#	endif

	/* this routine drives OpacityAdd1Subshell to put in total opacities for all shells*/

	/*begin sanity check */
	assert (ipZ >=0 );
	assert( ipZ < LIMELM );

	/* do not include the hydrogenic level */
	for( nion=0; nion < ipZ; nion++ )
	{
		if( xIonFracs[nion+1][ipZ] > 0. )
		{
			/*start with static opacities, then go to volatile*/
			
			chStat = 's';
			/* number of bound electrons */
			for( nshell=0; nshell < nsShellsCom.nsShells[nion][ipZ]; nshell++ )
			{
				/* highest shell will be volatile*/
				if( nshell== nsShellsCom.nsShells[nion][ipZ]-1 )
					chStat = 'v';
				/* set lower and upper limits to this range */
				low = opac.ipElement[0][nshell][nion][ipZ];
				ipHi = opac.ipElement[1][nshell][nion][ipZ];
				ipop = opac.ipElement[2][nshell][nion][ipZ];
				/* OpacityAdd1Subshell will not do anything if static opacities do not need to be reset*/
				OpacityAdd1Subshell(ipop,low,ipHi,xIonFracs[nion+1][ipZ] , chStat );
			}
		}
	}

	/* now do hydrogenic, but only if present 
	 * test for ipZ+1 in case atom present but not ion, test is whether the
	 * abundance of the recombined species is present */
	if( xIonFracs[ipZ+1][ipZ] > 0. )
	{
		/* do ground first, then all excited states */
		n = ipH1s;
		/* abundance of recombined species, which can be zero if no ion present */
		abundance = hydro.hn[ipZ][n]*xIonFracs[ipZ+2][ipZ];
		if( abundance == 0. )
		{
			abundance = hydro.hn[ipZ][n]*xIonFracs[ipZ+2][ipZ];
		}
		else 
		{
			/* no ionized species, assume everything in ground */
			abundance = xIonFracs[ipZ+1][ipZ];
		}

		/* use computed opacities and departure coef for level */
		OpacityAdd1SubshellInduc(
			opac.ipHOpac[ipZ][n],
			nh.ipHn[ipZ][n],
			/* the upper limit to the integration, 
			 * ground opacity goes up to the high energy limit of code*/
			rfield.nflux,
			hydro.hn[ipZ][n]*xIonFracs[ipZ+2][ipZ],
			/* departure coef, volatile opac, always reevaluate */
			hydro.hbn[ipZ][n] , 'v' );

		/* do excited levvels,
		 * this loop only if upper levels have finite population*/
		if( hydro.hn[ipZ][3]*xIonFracs[ipZ+2][ipZ] > 0. )
		{
			/* always want to evaluate all opacities for n=3, 4 */
			for( n=ipH2s; n < 5; n++ )
			{
				/* this is for lower levels of H, uses computed opacities */
				/* this is departure coef for level */
				OpacityAdd1SubshellInduc(
					opac.ipHOpac[ipZ][n],
					nh.ipHn[ipZ][n],
					/* the high energy bound of excited states is the 
					 * edge of the Lyman continuum */
					nh.ipHn[ipZ][ipH1s],
					hydro.hn[ipZ][n]*xIonFracs[ipZ+2][ipZ],
					/* departure coef, volitile opacities */
					hydro.hbn[ipZ][n] , 'v' );
			}

			/* NHPLPHOT lowest level where hydrogenic nu^-3 photo cross sect used*/
			for( n=5; n < MIN2( nHLevels[ipZ], NHPLPHOT); n++ )
			{
				/* this is for lower levels of H, uses computed opacities */
				/* this is departure coef for level */
				OpacityAdd1SubshellInduc(
					opac.ipHOpac[ipZ][n],
					nh.ipHn[ipZ][n],
					/* the high energy bound of excited states is the 
					 * edge of the Lyman continuum */
					nh.ipHn[ipZ][ipH1s],
					hydro.hn[ipZ][n]*xIonFracs[ipZ+2][ipZ],
					/* departure coef, static opacities */
					hydro.hbn[ipZ][n] , 's' );
			}

			/* now do upper levels that are only powerlaws */
			for( n=NHPLPHOT; n <= nHLevels[ipZ]; n++ )
			{
				/* this is for highly excited states, where simple
				 * power laws are used */
				opinPL(
					n,
					ipZ,
					hydro.hn[ipZ][n]*xIonFracs[ipZ+2][ipZ],
					hydro.hbn[ipZ][n]);
			}
		}
	}


#	ifdef DEBUG_FUN
	fputs( " <->OpacityAdd1Element()\n", debug_fp );
#	endif
	return;
}

/*opinPL add opacity of individual species, with stimulated emission, for power law opacity */
static void opinPL(
	long int n,	/* quantum number of level on C not physical scale, 
					 * must be less than NHYDRO_MAX_LEVEL, dim of STH */
  long int ipZ,/* charge, ipZ=0 for h, 1 for he, etc */
  float abundance, 
  double b)
{
	long int i, 
	  ipHi;
	float binv, 
	  csThresh;

#	ifdef DEBUG_FUN
	fputs( "<+>opinPL()\n", debug_fp );
#	endif

	/*add opacity of individual species, with stimulated emission, for power law opacity
	 *abundance is the density of the lower level (cm^-3)
	 *b is its departure coefficient, can be zero, is double precision */

	/* do nothing if abundance is zero, 
	 * or if no need to redo static opacities */
	if( abundance <= 0. || !opac.lgRedoStatic )
	{ 
#		ifdef DEBUG_FUN
		fputs( " <->opinPL()\n", debug_fp );
#		endif
		return;
	}

	/* following is numer of levels known - in HydroPhoCS.STH, where threshold
	 * photoionization cross sections are stored */
	assert( n < NHYDRO_MAX_LEVEL );

	csThresh = HydroPhoCS.STH[n]*rfield.anu3[nh.ipHn[ipZ][n]-1]/
	  POW2(ipZ+1.f);

	csThresh *= abundance;

	/* pick highest continuum to fill in */
	if( n == ipH1s )
	{
		/* for Lyman continuum, go up to highest defined boltzmann factor */
		ipHi = rfield.nflux;
	}
	else
	{
		/* for excited states go either up to Lyman continuum or highest
		 * defined Boltzmann factor */
		ipHi = MIN2(rfield.nflux,nh.ipHn[ipZ][ipH1s]);
	}

	/* b is dep coef, lgFluor is turned off with 'no indcued' command */
	if( (b > 1e-35 && fluors.lgFluor) && HPhoto.lgHInducImp )
	{
		binv = (float)(1./b);
		for( i=nh.ipHn[ipZ][n]-1; i < ipHi; i++ )
		{
			opac.OpacStatic[i] += csThresh/rfield.anu3[i]*
				MAX2(0.f,1.f- rfield.ContBoltz[i]*binv);
		}
	}
	else
	{
		/* B is the departure coef, which can go to zero, neglect stim emin */
		for( i=nh.ipHn[ipZ][n]-1; i < ipHi; i++ )
		{
			opac.OpacStatic[i] += csThresh/rfield.anu3[i];
		}
	}

#	ifdef DEBUG_FUN
	fputs( " <->opinPL()\n", debug_fp );
#	endif
	return;
}

