/*ParseAtomHLike parse information from the hydrogen command line */
#include <string.h>
#include "cddefines.h"
#include "taulines.h"
#include "hydrogenic.h"
#include "dampon.h"
#include "physok.h"
#include "redis.h"
#include "rfield.h"
#include "thlo.h"
#include "parse.h"

void ParseAtomHLike(char *chCard )
{
	int lgEOL;
	long int i , 
		nh ,
		ipZ;

#	ifdef DEBUG_FUN
	fputs( "<+>ParseAtomHLike()\n", debug_fp );
#	endif


	if( lgMatch("COLL",chCard) )
	{
		/* option to turn hydrogen collisions off, all are
		 * set to 1 in zero.  command can accept only one option at a time */
		if( lgMatch("EXCI",chCard) )
		{
			/* turn off collisional excitation */
			hydro.HColExcOn = 0.;
			physok.lgPhysOK = FALSE;
		}
		else if( lgMatch("IONI",chCard) )
		{
			/* turn off collisional ionization */
			hydro.HColIonOn = 0.;
			physok.lgPhysOK = FALSE;
		}
		else if( lgMatch("2S2P",chCard) || lgMatch("2P2S",chCard) )
		{
			/* turn off 2s - 2p collisions */
			hydro.lgC2psOn = FALSE;
			physok.lgPhysOK = FALSE;
		}
		else
		{
			/* turn everything off, since no keyword given */
			hydro.HColExcOn = 0.;
			hydro.HColIonOn = 0.;
			hydro.lgC2psOn = FALSE;
			physok.lgPhysOK = FALSE;
		}
	}

	else if( lgMatch("DAMP",chCard) )
	{
		/* turn off absorption due to lyalpha damping wings */
		dampon.DampOnFac = 0.;
	}

	else if( lgMatch("EMIS",chCard) )
	{
		/* hydrogenic emission lines will be emissivity not intensity or luminosity */
		hydro.lgHydEmiss = TRUE;
	}

	else if( lgMatch("LEVE",chCard) )
	{
		/* number of levels for hydrogen at, 2s is this plus one */
		/* there are two options here,
		 * when keyword ELEMENT appears, scan off element name and change levels only for
		 * that one.
		 * when there is no ELEMENT then set all nhlevels to same number */

		/* lgHydroMalloc is FALSE at start of calculation, set true when space 
		 * allocated for the hydrogen lines.  Once done we must ignore all 
		 * future changes in the number of levels */
		if( !lgHydroMalloc )
		{
			i = 5;
			nh = (long int)FFmtRead(chCard,&i,76,&lgEOL);

			if( lgEOL )
			{
				/* must be a number, or a key, either large (or limit) or small
				 * if no number but LARGER or SMALL then set to preset number */
				if( lgMatch("LARG",chCard) )
				{
					/* there is no real limit, but this seems pretty big */
					nh = 50;
				}

				/* this is small or compact keyword */
				else if( lgMatch("SMAL",chCard) || lgMatch("COMP",chCard) )
				{
					/* a small atom is the same as c84 */
					nh = 10;
				}

				/* this is limit keyword, sets to largest possible value */
				else if( lgMatch("LIMI",chCard) )
				{
					nh = NHYDRO_MAX_LEVEL-1;
				}

				else
				{
					/* punch out if no number */
					NoNumb(chCard);
				}

			}

			if( nh < 4 )
			{
				fprintf( ioQQQ, " This would be too few levels.\n" );
				puts( "[Stop in ParseAtomHLike]" );
				cdEXIT(1);
			}

			if( nh < hydro.nHTopOff )
			{
				fprintf( ioQQQ, " Not possible to set nhlvl to less than top off=%4li\n", 
				  hydro.nHTopOff );
				fprintf( ioQQQ, " change nHTopOff first with ATOM H-LIKE TOPOFF command\n");
				puts( "[Stop in ParseAtomHLike]" );
				cdEXIT(1);
			}

			if( nh > NHYDRO_MAX_LEVEL-1 )
			{
				fprintf( ioQQQ, " Not possible to set nhlvl to >NHYDRO_MAX_LEVEL %li\n", 
				  NHYDRO_MAX_LEVEL );
				fprintf( ioQQQ, " change NHYDRO_MAX_LEVEL\n");
				puts( "[Stop in ParseAtomHLike]" );
				cdEXIT(1);
			}

			/* this is check that alpha transition of highest level is within energy
			 * bounds of continuum */
			if( 2. / POW3((double)nh) < rfield.emm )
			{
				fprintf( ioQQQ, " Not possible to set nHLevels[ipHYDROGEN] to such a high value, since "
					"alpha transition not within energy bounds of code\n");

				fprintf( ioQQQ, " lowest energy is %e and corresponding highest level is %li\n" ,
					rfield.emm, (long)pow(2./rfield.emm, 0.3333) );
				puts( "[Stop in ParseAtomHLike]" );
				cdEXIT(1);
			}

			/* we now have the desired number of levels, and it has been fully qualified. 
			 * now check whether ELEMENT keyword is on line, if not then set all to
			 * the number, if so then only element in use */
			if( lgMatch("ELEM",chCard) )
			{
				/* check which element is on the line, ipZ is atomic number on C scale */
				GetElem(chCard,&ipZ);
				/* only set number of levels for this one element */
				nHLevels[ipZ] = nh;
			}
			else
			{
				/* keyword element did not appear, so set all h-like species to this number */
				for( i=0; i<LIMELM; ++i )
				{
					nHLevels[i] = nh;
				}
			}

		}

	}

	else if( lgMatch("LOWE",chCard) )
	{
		/* set lowest temp for departure coef, default is 1000 */
		i = 5;
		thlo.HydTempLimit = (float)FFmtRead(chCard,&i,76,&lgEOL);
		if( thlo.HydTempLimit <= 10. )
			thlo.HydTempLimit = (float)pow(10.,thlo.HydTempLimit);
	}

	else if( lgMatch("REDI",chCard) )
	{
		/* set redistribution function either wind, incompl (default), or compl */
		if( lgMatch(" COM",chCard) )
		{
			/* complete redistribution for la and all others */
			strcpy( redis.chRedis, "COMP" );
		}

		else if( lgMatch("INCO",chCard) )
		{
			/* incomplete redistribution for all resonance lines */
			strcpy( redis.chRedis, "INCO" );
		}

		else
		{
			fprintf( ioQQQ, " MUST be either COMPLETE or INCOMPL\n" );
			puts( "[Stop in ParseAtomHLike]" );
			cdEXIT(1);
		}
	}

	else if( lgMatch("TOPO",chCard) )
	{
		/* set lowest level for hydrogen atom top off */
		i = 5;
		hydro.nHTopOff = (long int)FFmtRead(chCard,&i,76,&lgEOL);

		if( hydro.nHTopOff > nHLevels[ipHYDROGEN] )
		{
			fprintf( ioQQQ, " Not possible to set nHTopOff to more than nhlvl=%4ld\n", 
			  nHLevels[ipHYDROGEN] );
			fprintf( ioQQQ, " change nhlvl first with ATOM H-LIKE LEVELS command\n");
			puts( "[Stop in ParseAtomHLike]" );
			cdEXIT(1);
		}

		if( lgEOL )
		{
			NoNumb(chCard);
		}

		if( lgMatch(" ADD",chCard) )
		{
			/* add extra part */
			strcpy( hydro.chHTopType, " add" );
		}
		else if( lgMatch("SCAL",chCard) )
		{
			/* use scale factors */
			strcpy( hydro.chHTopType, "scal" );
		}
		else
		{
			/* this is the default */
			strcpy( hydro.chHTopType, "scal" );
		}
	}

	/* which type of matrix solution, populations, lowt, departure coef */
	else if( lgMatch("MATR",chCard) )
	{
		if( lgMatch("POPU",chCard) )
		{
			strcpy( hydro.chTypeAtom , "POPU" );
		}
		else if( lgMatch("LOWT",chCard) )
		{
			strcpy( hydro.chTypeAtom , "LOWT" );
		}
		else if( lgMatch("DEPA",chCard) )
		{
			strcpy( hydro.chTypeAtom , "DEPA" );
		}
		else
		{
			fprintf( ioQQQ, " There should have been a keyword on the MATRIX option.\n");
			fprintf( ioQQQ, " The keywords are POPUlations, LOWTe, and DEPArture coef.\n");
			fprintf( ioQQQ, " Stop in ParseAtomHLike\n" );
			puts( "[Stop in ParseAtomHLike]" );
			cdEXIT(1);
		}
	}

	else
	{
		fprintf( ioQQQ, " There should have been a keyword - Stop in ParseAtomHLike\n" );
		puts( "[Stop in ParseAtomHLike]" );
		cdEXIT(1);
	}

#	ifdef DEBUG_FUN
	fputs( " <->ParseAtomHLike()\n", debug_fp );
#	endif
	return;
}

