/* 
 * a set of standard routines that are widely used across the code for various
 * houeskeeping chores.  These do not do any physics, and are unlikely to
 * change over time.
 *
 * the prototypes are in cddefines.h and so are automatically picked up by all routines 
 *
 */
/*FFmtRead scan input line for free format number */
/*e2 second exponential integral */
/*caps convert input command line (through eol) to ALL CAPS */
/*ShowMe produce request to send information to GJF after a crash */
/*AnuUnit produce continuum energy in arbitrary units */
/*cap4 convert first 4 char of input line chLab into chCAP all in caps, null termination */
/*insane set flag saying that insanity has occurred */
/*lgMatch determine whether match to a keyword occurs on command line,
 * return value is 0 if no match, and position of match within string if hit */
/*fudge enter fudge factors, or some arbitrary number, with fudge command*/
/*GetElem scans line image, finds element. returns atomic number j, on C scale */
/*GetQuote get any name between double quotes off command line
 * return string as chLabel, is null terminated */
/*NoNumb general error handler for no numbers on input line */
/*sexp safe exponential function */
/*TestCode set flag saying that test code is in place */
/*fixit - say that code needs to be fixed */
/*broken set flag saying that the code is broken, */
/*dprintf is a debug print routine that was provided by Peter Teuben,
 * as a component from his NEMO package.  It offers run-time specification
 * of the level of debugging */
/*qg32 32 point Gaussian quadrature, original fortran given to Gary F by Jim Lattimer */
#include <ctype.h>
#include <stdarg.h>	/* ANSI variable arg macros */
#include <string.h>
#include "cddefines.h"
#include "cddrive.h"
#include "testit.h"
#include "insanity.h"
#include "fudgec.h"
#include "broke.h"
#include "trace.h"
#include "input.h"
#include "elementnames.h"
#include "physconst.h"
#include "pnunit.h"
#include "date.h"
#include "warnings.h"
#include "converge.h"

/*AnuUnit produce continuum energy in arbitrary units */
double AnuUnit(float energy)
/*static double AnuUnit(long int ip)*/
{
	double AnuUnit_v;

#	ifdef DEBUG_FUN
	fputs( "<+>AnuUnit()\n", debug_fp );
#	endif

	if( energy <=0. )
	{
		/* this is insanity */
		AnuUnit_v = 0.;
	}
	else if( strcmp(pnunit.chConPunEnr[pnunit.ipConPun],"ryd ") == 0 )
	{
		/* energy in Ryd */
		AnuUnit_v = energy;
	}
	else if( strcmp(pnunit.chConPunEnr[pnunit.ipConPun],"micr") == 0 )
	{
		/* wavelength in microns */
		AnuUnit_v = RYDLAM/energy*1e-4;
	}
	else if( strcmp(pnunit.chConPunEnr[pnunit.ipConPun]," kev") == 0 )
	{
		/* energy in keV */
		AnuUnit_v = energy*EVRYD*1.e-3;
	}
	else if( strcmp(pnunit.chConPunEnr[pnunit.ipConPun]," ev ") == 0 )
	{
		/* energy in eV */
		AnuUnit_v = energy*EVRYD;
	}
	else if( strcmp(pnunit.chConPunEnr[pnunit.ipConPun],"angs") == 0 )
	{
		/* wavelength in Angstroms */
		AnuUnit_v = RYDLAM/energy;
	}
	else
	{
		fprintf( ioQQQ, " insane units in AnuUnit =%4.4s\n", 
		  pnunit.chConPunEnr[pnunit.ipConPun] );
		puts( "[Stop in AnuUnit]" );
		cdEXIT(1);exit(1);
	}

#	ifdef DEBUG_FUN
	fputs( " <->AnuUnit()\n", debug_fp );
#	endif
	return( AnuUnit_v );
}

/*ShowMe produce request to send information to GJF after a crash */
void ShowMe(void)
{

#	ifdef DEBUG_FUN
	fputs( "<+>ShowMe()\n", debug_fp );
#	endif

	/* print info if output unit is defined */
	if( ioQQQ != NULL )
	{
		fprintf( ioQQQ, "\n\n" );
		fprintf( ioQQQ, "\n\n" );
		fprintf( ioQQQ, "           vvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvv \n" );
		fprintf( ioQQQ, "          > Something bad has happened.       <\n" );
		fprintf( ioQQQ, "          > Please show this to Gary Ferland. <\n" );
		fprintf( ioQQQ, "          > email:  gary@cloud9.pa.uky.edu    <\n" );
		fprintf( ioQQQ, "          > Please send all following info:   <\n" );
		fprintf( ioQQQ, "           ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^ \n" );
		fprintf( ioQQQ, "\n\n" );

		fprintf( ioQQQ, " Cloudy version number is %7.7s\n", 
			date.chVersion );

		fprintf( ioQQQ, "%5ld warnings,%3ld cautions,%3ld temperature failures.  Messages follow.\n", 
		  warnings.nwarn, warnings.ncaun, conv.nTeFail );

		/* print the warnings first */
		cdWarnings(ioQQQ);

		/* now print the cautions */
		cdCautions(ioQQQ);

		/* now output the commands */
		cdPrintCommands(ioQQQ);
	}

#	ifdef DEBUG_FUN
	fputs( " <->ShowMe()\n", debug_fp );
#	endif
	return;
}

/*cap4 convert first 4 char of input line chLab into chCAP all in caps, null termination */
void cap4(
		char *chCAP ,	/* output string, cap'd first 4 char of chLab, */
	                  /* with null terminating */
		char *chLab)	/* input string ending with eol*/
{
	long int /*chr,*/ 
	  i ;

#	ifdef DEBUG_FUN
	fputs( "<+>cap4()\n", debug_fp );
#	endif


	/* convert 4 character string in chLab to ALL CAPS in chCAP */
	for( i=0; i < 4; i++ )
	{
		/* toupper is function in ctype that converts to upper case */
		chCAP[i] = (char)toupper( chLab[i] );
	}

	/* now end string with eol */
	chCAP[4] = '\0';


#	ifdef DEBUG_FUN
	fputs( " <->cap4()\n", debug_fp );
#	endif
	return;
}

/*caps convert input command line (through eol) to ALL CAPS */
void caps(char *chCard )
{
	long int i;

#	ifdef DEBUG_FUN
	fputs( "<+>caps()\n", debug_fp );
#	endif

	/* convert full character string in chCard to ALL CAPS */
	i = 0;
	while( chCard[i]!= '\0' )
	{
		chCard[i] = (char)toupper( chCard[i] );
		++i;
	}


#	ifdef DEBUG_FUN
	fputs( " <->caps()\n", debug_fp );
#	endif
	return;
}

/*e2 second exponential integral */
double e2(double t, 
	  double tln)
{
	double e2_v, 
	  remain;

#	ifdef DEBUG_FUN
	fputs( "<+>e2()\n", debug_fp );
#	endif

	/* fit of second exponential integral;
	 * T is optical depth, and TLN is EXP(-t)
	 * */
	if( t < 0.3 )
	{
		remain = (1.998069357 + t*(66.4037741 + t*107.2041376))/(1. + 
		  t*(37.4009646 + t*105.0388805));

	}
	else if( t < 20. )
	{
		remain = (1.823707708 + t*2.395042899)/(1. + t*(2.488885899 - 
		  t*0.00430538));

	}
	else
	{
		remain = 1.;
	}

	e2_v = remain*tln/(2. + t);


#	ifdef DEBUG_FUN
	fputs( " <->e2()\n", debug_fp );
#	endif
	return( e2_v );
}

/*ee1 first exponential integral */
double ee1(double x)
{
	double ans, 
	  bot, 
	  top;
	static double a[6]={-.57721566,.99999193,-.24991055,.05519968,-.00976004,
	  .00107857};
	static double b[4]={8.5733287401,18.0590169730,8.6347608925,.2677737343};
	static double c[4]={9.5733223454,25.6329561486,21.0996530827,3.9584969228};

#	ifdef DEBUG_FUN
	fputs( "<+>ee1()\n", debug_fp );
#	endif

	/* computes the exponential integral E1(x),
	 * from abramowitz and stegun
	 * returns error condition for negative argument,
	 * returns zero when limit on sexp exceeded
	 * */

	/* error - this does not do complex numbers */
	if( x <= 0 )
	{
		fprintf( ioQQQ, " negative argument in function ee1, x<0\n" );
		puts( "[Stop in ee1]" );
		cdEXIT(1);exit(1);
	}

	/* branch for x less than 1 */
	else if( x < 1. )
	{
		/* abs. accuracy better than 2e-7 */
/* 		ans = a[0] + a[1]*x + a[2]*x*x + a[3]*powi(x,3) + a[4]*powi(x,4) +  */
/* 		  a[5]*powi(x,5) - log(x); */
		ans = ((((a[5]*x + a[4])*x + a[3])*x + a[2])*x + a[1])*x + a[0] - log(x);
	}

	/* branch for x greater than, or equal to, one */
	else
	{
		/* abs. accuracy better than 2e-8 */
/* 		top = powi(x,4) + b[0]*powi(x,3) + b[1]*x*x + b[2]*x + b[3]; */
		top = (((x + b[0])*x + b[1])*x + b[2])*x + b[3];
/* 		bot = powi(x,4) + c[0]*powi(x,3) + c[1]*x*x + c[2]*x + c[3]; */
		bot = (((x + c[0])*x + c[1])*x + c[2])*x + c[3];
		/* 
		 *>>>chng 98 dec 17
		 * Jason's original implementation did not require the exp since
		 * the routines that used this knew if was not present.  
		 * put in for safely.
		 
		ans = top/bot/x;*/
		ans = top/bot/x*sexp(x);
	}

#	ifdef DEBUG_FUN
	fputs( " <->ee1()\n", debug_fp );
#	endif
	return( ans );
}

/*FFmtRead scan input line for free format number */

/********************************************************************
 * start of nxtchr, used only by FFmtRead                           *
 ********************************************************************/
static void nxtchr(char *chr, 
	  long int *ichr, 
	  long int ipnt, 
	  char *chCard, 
	  long int last, 
	  int *lgEOL)
{

#	ifdef DEBUG_FUN
	fputs( "<+>nxtchr()\n", debug_fp );
#	endif

	/* read character chCard(IPNT:IPNT), return charac as CHR, octal as ICHR
	 * return lgEOL=.TRUE. if END reached
	 * */
	*chr = chCard[ipnt - 1];
	*ichr = (*chr);

	/* 
	 * check for end of line, either last as passed to FFmtRead,
	 * or (most likely) the end of line character 
	 */
	if( ipnt >= last || *chr=='\0' )
	{
		*lgEOL = TRUE;
	}
	else
	{
		*lgEOL = FALSE;
	}


#	ifdef DEBUG_FUN
	fputs( " <->nxtchr()\n", debug_fp );
#	endif
	return;
}

/********************************************************************
 * start of FFmtRead                                                *
 ********************************************************************/
double FFmtRead(char *chCard, 
	  long int *ipnt, 
	  long int last, 
	  int *lgEOL)
{
	int lgNFound, 
	  lgNumber;
	char chr;
	long int ichr, 
	  l1, 
	  l2;
	double FFmtRead_v, 
	  expn, 
	  PM1, 
	  value;

#	ifdef DEBUG_FUN
	fputs( "<+>FFmtRead()\n", debug_fp );
#	endif

	/* FFmtRead= first number encountered after column IPNT in card image */

	/* ipnt is pointer to position within array, must be in range 1 to 
	 * longest possible line image */
	assert(*ipnt > 0 );
	assert( *ipnt < 132 );

	l1 = '0';
	l2 = '9';

L_999:
	lgNumber = FALSE;
	lgNFound = FALSE;
	PM1 = 1.;
	value = 0.;
	FFmtRead_v = 0.;
	/* have we hit the end of line? */
	nxtchr(&chr,&ichr,*ipnt,chCard,last,lgEOL);
	if( *lgEOL )
	{ 
#		ifdef DEBUG_FUN
		fputs( " <->FFmtRead()\n", debug_fp );
#		endif
		return( FFmtRead_v );
	}

	/*************************************************
	 * find start of number                          *
	 *************************************************/
L_4:
	if( chr == '.' )
		goto L_5;

	if( (ichr >= l1) && (ichr <= l2) )
		goto L_1;
	*ipnt += 1;
	nxtchr(&chr,&ichr,*ipnt,chCard,last,lgEOL);
	if( *lgEOL )
	{ 
#		ifdef DEBUG_FUN
		fputs( " <->FFmtRead()\n", debug_fp );
#		endif
		return( FFmtRead_v );
	}
	goto L_4;

	/*************************************************
	 * found start of number, check if neg           *
	 *************************************************/
L_1:
	/* only do this test if we are deep in the line */
	if( *ipnt>1 )
	{
		if( chCard[*ipnt - 2] == '-' )
		  PM1 = -1.;
		lgNFound = TRUE;
	}

	/*************************************************
	 * find numerical value                          *
	 *************************************************/
L_6:
	value = 10.*value + (float)(labs(ichr-l1));
	lgNumber = TRUE;
L_7:
	*ipnt += 1;
	nxtchr(&chr,&ichr,*ipnt,chCard,last,lgEOL);
	if( *lgEOL )
	{
		FFmtRead_v = value*PM1;
		/* will be encountered on next call */
		*lgEOL = FALSE;
		
#		ifdef DEBUG_FUN
		fputs( " <->FFmtRead()\n", debug_fp );
#		endif
		return( FFmtRead_v );
	}
	if( chr == ',' )
		goto L_7;
	if( (ichr >= l1) && (ichr <= l2) )
		goto L_6;
	/***************************************************/
	if( *ipnt > 0 )
	{
		if( chCard[*ipnt - 1] != '.' )
		{
			FFmtRead_v = value*PM1;
			
#			ifdef DEBUG_FUN
			fputs( " <->FFmtRead()\n", debug_fp );
#			endif
			return( FFmtRead_v );
		}
	}

	/* decimal point encountered */
L_5:
	expn = 1.;
	if( *ipnt > 1 )
	{
		if( chCard[*ipnt - 2] == '-' )
			PM1 = -1.;
	}

	/*****************************************************/
L_3:
	*ipnt += 1;
	nxtchr(&chr,&ichr,*ipnt,chCard,last,lgEOL);
	if( chr == ',' )
		goto L_3;
	if( (*lgEOL || ichr < l1) || ichr > l2 )
	{
		/* following fails if no number found (only . on line) */
		if( !lgNFound )
			goto L_999;
		FFmtRead_v = value*PM1;
		if( lgNumber )
			*lgEOL = FALSE;
		
#		ifdef DEBUG_FUN
		fputs( " <->FFmtRead()\n", debug_fp );
#		endif
		return( FFmtRead_v );
	}
	expn *= 0.1;
	lgNFound = TRUE;
	lgNumber = TRUE;
	value += expn*(float)(ichr-l1);
	goto L_3;
	/*****************************************************/
}

/*insane set flag saying that insanity has occurred */
void insane(void)
{

#	ifdef DEBUG_FUN
	fputs( "<+>insane()\n", debug_fp );
#	endif

	/* set flag lgInsane to true
	 * called if somethine impossible happened
	 * a comment will be printed after the calculation, along with
	 * a request to show me
	 * */
	insanity.lgInsane = TRUE;

#	ifdef DEBUG_FUN
	fputs( " <->insane()\n", debug_fp );
#	endif
	return;
}

static int istrstr( char *str, char *substr )	/* return the int location of substr in str */
/*char *str, *substr;*/
{
	char *s;

	if( substr==NULL || *substr=='\0' )
		return( (int)strlen(str)+1 ); /*from Chuck Ditter of Infinity Graphics,6/93*/
	else if( (s=strstr(str,substr)) == NULL )
		return( 0 );	/* not present */
	return( (int)(s-str) + 1 );
}

/*lgMatch determine whether match to a keyword occurs on command line,
 * return value is 0 if no match, and position of match within string if hit */
int lgMatch(char *chKey, 
	  char *chCard)
{
	int lgMatch_v;
	long int i;

#	ifdef DEBUG_FUN
	fputs( "<+>lgMatch()\n", debug_fp );
#	endif

	/* cobalt blue routine that returns 0 if string is not in other,
	 * else returns position of chKey within chCard */
	i = istrstr(chCard,chKey);
	if( i == 0 )
	{
		/* did not find match, returned false */
		lgMatch_v = FALSE;
	}
	else
	{
		/* value was position with chCard, return this to lgMatch */
		lgMatch_v = (int)i;
	}

#	ifdef DEBUG_FUN
	fputs( " <->lgMatch()\n", debug_fp );
#	endif
	return( lgMatch_v );
}

/* fudge enter fudge factors, or some arbitrary number, with fudge command
 * other sections of the code access these numbers by calling fudge
 * fudge(0) returns the first number that was entered
 * prototype for this function is in cddefines.h so that it can be used without
 * declarations */
double fudge(long int ipnt)
{
	double fudge_v;

#	ifdef DEBUG_FUN
	fputs( "<+>fudge()\n", debug_fp );
#	endif

	if( ipnt >= fudgec.nfudge )
	{
		fprintf( ioQQQ, " FUDGE factor not entered for array number %3ld\n", 
		  ipnt );
		puts( "[Stop in fudge]" );
		cdEXIT(1);exit(1);
	}
	else
	{
		fudge_v = fudgec.fudgea[ipnt];
	}

#	ifdef DEBUG_FUN
	fputs( " <->fudge()\n", debug_fp );
#	endif
	return( fudge_v );
}

/*GetElem scans line image, finds element. returns atomic number j, on C scale */
void GetElem(char *chCard,
  long int *j)
{
	int lgFound;
	long int i;

#	ifdef DEBUG_FUN
	fputs( "<+>GetElem()\n", debug_fp );
#	endif


	/* find which element */
	i = 1;
	lgFound = FALSE;
	while( i <= LIMELM && !lgFound )
	{
		if( lgMatch((char*)elementnames.chElementNameShort[i-1],chCard) )
		{
			/* IEL is atomic number of element, is used below */
			*j = i;
			lgFound = TRUE;
		}
		++i;
	}

	if( !lgFound )
	{
		fprintf( ioQQQ, " GetElem did not find an element on the following line:\n" );
		fprintf( ioQQQ, "%s\n", chCard );
		puts( "[Stop in getelem]" );
		cdEXIT(1);
	}
	/* get back to C scale, the expected return */
	--*j;

#	ifdef DEBUG_FUN
	fputs( " <->GetElem()\n", debug_fp );
#	endif
	return;
}

/* GetQuote get any name between double quotes off command line
 * return string as chLabel, is null terminated */
void GetQuote(char *chLabel,	/* we will generate a label and stuff it here */
	      char *chCard )	/* local capd line, we will blank this out */	
{
	char *i0,       /* pointer to first " */
	  *i1,          /* pointer to second ", name is in between */
	  *iLoc;        /* pointer to first " in local version of card in calling routine */
	size_t len;

#	ifdef DEBUG_FUN
	fputs( "<+>GetQuote()\n", debug_fp );
#	endif

	/*label within quotes returned within chLabel
	 *label in line image is set to blanks when complete */

	/* find start of string, string must begin and end with double quotes */
	/* get pointer to first quote */
	i0 = strchr( input.chOrgCard,'\"' );

	if( i0 != NULL ) {
		/* get pointer to next quote */
		/*i1 = strrchr( input.chOrgCard,'\"' );*/
		i1 = strchr( i0+1 ,'\"' );
	}
	else {
		i1 = NULL;
	}

	/* check that pointers were not NULL */
	/* >>chng 00 apr 27, check for i0 and i1 equal not needed anymore, by PvH */
	if( i0 == NULL || i1 == NULL )
	{
		fprintf( ioQQQ, 
			" A filename or label must be specified within double quotes, but no quotes were encountered on this command.\n" );
		fprintf( ioQQQ, " Name must be surrounded by exactly two double quotes, like \"name.txt\". \n" );
		fprintf( ioQQQ, " The line image follows:\n" );
		fprintf( ioQQQ, " %s\n", input.chOrgCard);
		fprintf( ioQQQ, " Sorry\n" );
		puts( "[Stop in getquote]" );
		cdEXIT(1);
	}

	/* now copy the text inbetween quotes */
	len = (size_t)(i1-i0-1);
	strncpy(chLabel,i0+1,len);
	/* strncpy doesn't terminate the label */
	chLabel[len] = '\0';

	/* get pointer to first quote in local copy of line image in calling routine */
	iLoc = strchr( chCard ,'\"' );
	if( iLoc == NULL )
	{
		fprintf( ioQQQ, " Insanity in GetQuote - line image follows:\n" );
		fprintf( ioQQQ, " %s\n", input.chOrgCard);
		puts( "[Stop in getquote]" );
		cdEXIT(1);
	}

	/* >>chng 97 jul 01, blank out label once finished, to not be picked up later */
	/* >>chng 00 apr 27, erase quotes as well, so that we can find second label, by PvH */
	while( i0 <= i1 )
	{
		*i0 = ' ';
		*iLoc = ' ';
		++i0;
		++iLoc;
	}

#	ifdef DEBUG_FUN
	fputs( " <->GetQuote()\n", debug_fp );
#	endif
	return;
}

/* powi.c - calc x^n, where n is an integer! */

/* Very slightly modified version of power() from Computer Language, Sept. 86,
	pg 87, by Jon Snader (who extracted the binary algorithm from Donald Knuth,
	"The Art of Computer Programming", vol 2, 1969).
	powi() will only be called when an exponentiation with an integer
	exponent is performed, thus tests & code for fractional exponents were 
	removed.	-- cw lightfoot, COBALT BLUE
  COBALT BLUE has NO copyright on the functions in this file.  The source may 
  be considered public domain.
 */
#define IS_ODD(j)	((j) & 1 )

/* this will only define powi if this is not an alpha running the native os */
#if !defined(__alpha) || defined(__linux)

/* #ifndef __alpha   this fcn is on an alpha, only defn if not alpha */

double powi( double x, long int n )	/* returns:  x^n */
/* x;	 base */
/* n;	 exponent */
{
	double p;	/* holds partial product */

	if( x == 0 )
		return( 0. );

	/* test for negative exponent */
	if( n < 0 )
	{	
		n = -n;
		x = 1/x;
	}

	p = IS_ODD(n) ? x : 1;	/* test & set zero power */

	while( n >>= 1 ){	/* now do the other powers */
		x *= x;			/* sq previous power of x */
		if( IS_ODD(n) )	/* if low order bit set */
			p *= x;		/*	then, multiply partial product by latest power of x */
		}

	return( p );
}

#	endif /*alpha unix*/

	/*		*		*		*		*/

long ipow( long m, long n )	/* returns:  m^n */
/* m;		 base */
/* n;		 exponent */
{
long p;	/* holds partial product */

if( m == 0 || (n < 0 && m > 1) )
	return( 0L );
	/* NOTE: negative exponent always results in 0 for integers!
	 * (except for the case when m==1 or -1) */

if( n < 0 ){	/* test for negative exponent */
	n = -n;
	m = 1/m;
	}

p = IS_ODD(n) ? m : 1;	/* test & set zero power */

while( n >>= 1 ){	/* now do the other powers */
	m *= m;			/* sq previous power of m */
	if( IS_ODD(n) )	/* if low order bit set */
		p *= m;		/*	then, multiply partial product by latest power of m */
	}

return( p );
}

/*PrintE82 - series of routines to mimic 1p, e8.2 fortran output */
/***********************************************************
 * contains the following sets of routines to get around   *
 * the MS C++ compilers unusual exponential output.        *
 * PrintEfmt <= much faster, no overhead with unix         *
 * PrintE93                                                *
 * PrintE82                                                *
 * PrintE71                                                *
 **********************************************************/

/**********************************************************/
/*
* Instead of printf'ing with %e or %.5e or whatever, call
* efmt("%e", val) and print the result with %s.  This lets
* us work around bugs in VS C 6.0.
*/
char *PrintEfmt(const char *fmt, double val /* , char *buf */) 
{
	static char buf[30]; /* or pass it in */

	/* create the string */
	sprintf(buf, fmt, val);

	/* we need to fix e in format if ms vs */
#	ifdef _MSC_VER
	{
		/* code to fix incorrect ms v e format.  works only for case where there is
		 * a leading space in the format - for formats like 7.1, 8.2, 9.3, 10.4, etc
		 * result will have 1 too many characters */
		char *ep , buf2[30];

		/* msvc behaves badly in different ways for positive vs negative sign vals,
		 * if val is positive must create a leading space */
		if( val >= 0.)
		{
			strcpy(buf2 , " " );
			strcat(buf2 , buf);
			strcpy( buf , buf2);
		}

		/* allow for both e and E formats */
		if ((ep = strchr(buf, 'e')) == NULL)
		{
			ep = strchr(buf, 'E');
		}

		/* ep can still be NULL if val is Inf or NaN */
		if (ep != NULL) 
		{
			/* move pointer two char past the e, to pick up the e and sign */
			ep += 2;

			/* terminate buf where the e is, *ep points to this location */
			*ep = '\0';

			/* skip next char, */
			++ep;

			/* copy resulting string to return string */
			strcat( buf, ep );
		}
	}
#	endif

	return buf;
}
/**********************************************************/
void PrintE82( FILE* ioOut, double value )
{
	double frac , xlog , xfloor , tvalue;
	int iExp;
	if( value < 0 )
	{
		fprintf(ioOut,"********");
	}
	else if( value == 0 )
	{
		fprintf(ioOut,"0.00E+00");
	}
	else
	{
		/* round number off for 8.2 format (not needed since can't be negative) */
		tvalue = value;
		xlog = log10( tvalue );
		xfloor = floor(xlog);
		/* this is now the fractional part */
		frac = tvalue*pow(10.,-xfloor);
		/*this is the possibly signed exponential part */
		iExp = (int)xfloor;
		if( frac>9.9945 )
		{
			frac /= 10.;
			iExp += 1 ;
		}
		/* print the fractional part*/
		fprintf(ioOut,"%.2f",frac);
		/* E for exponent */
		fprintf(ioOut,"E");
		/* if positive throw a + sign*/
		if(iExp>=0 )
		{
			fprintf(ioOut,"+");
		}
		fprintf(ioOut,"%.2d",iExp);
	}

	return;
}
/*
 *==============================================================================
 */
void PrintE71( FILE* ioOut, double value )
{
	double frac , xlog , xfloor , tvalue;
	int iExp;
	if( value < 0 )
	{
		fprintf(ioOut,"*******");
	}
	else if( value == 0 )
	{
		fprintf(ioOut,"0.0E+00");
	}
	else
	{
		/* round number off for 8.2 format (not needed since can't be negative) */
		tvalue = value;
		xlog = log10( tvalue );
		xfloor = floor(xlog);
		/* this is now the fractional part */
		frac = tvalue*pow(10.,-xfloor);
		/*this is the possibly signed exponential part */
		iExp = (int)xfloor;
		if( frac>9.9945 )
		{
			frac /= 10.;
			iExp += 1 ;
		}
		/* print the fractional part*/
		fprintf(ioOut,"%.1f",frac);
		/* E for exponent */
		fprintf(ioOut,"E");
		/* if positive throw a + sign*/
		if(iExp>=0 )
		{
			fprintf(ioOut,"+");
		}
		fprintf(ioOut,"%.2d",iExp);
	}

	return;
}
 
/*
 *==============================================================================
 */
void PrintE93( FILE* ioOut, double value )
{
	double frac , xlog , xfloor, tvalue;
	int iExp;
	if( value < 0 )
	{
		fprintf(ioOut,"*********");
	}
	else if( value == 0 )
	{
		fprintf(ioOut,"0.000E+00");
	}
	else
	{
		/* round number off for 9.3 format, neg numb not possilbe */
		tvalue = value;
		xlog = log10( tvalue );
		xfloor = floor(xlog);
		/* this is now the fractional part */
		frac = tvalue*pow(10.,-xfloor);
		/*this is the possibly signed exponential part */
		iExp = (int)xfloor;
		if( frac>9.99949 )
		{
			frac /= 10.;
			iExp += 1 ;
		}
		/* print the fractional part*/
		fprintf(ioOut,"%5.3f",frac);
		/* E for exponent */
		fprintf(ioOut,"E");
		/* if positive throw a + sign*/
		if(iExp>=0 )
		{
			fprintf(ioOut,"+");
		}
		fprintf(ioOut,"%.2d",iExp);
	}

	return;
}

/*NoNumb general error handler for no numbers on input line */
void NoNumb(char *chCard)
{

#	ifdef DEBUG_FUN
	fputs( "<+>NoNumb()\n", debug_fp );
#	endif

	/* general catch-all for no number when there should have been */
	fprintf( ioQQQ, " There is a problem on the following command line:\n" );
	fprintf( ioQQQ, " %s\n", chCard );
	fprintf( ioQQQ, " There should have been a number on this line.   Sorry.\n" );
	puts( "[Stop in nonumb]" );
	cdEXIT(1);

}

/*sexp safe exponential function */
double sexp(double x)
{
	double sexp_v;

#	ifdef DEBUG_FUN
	fputs( "<+>sexp()\n", debug_fp );
#	endif

	/* NB - the 84 in the following must be kept parallel with the 84 in boltgn,
	 * since level2 uses ContBoltz to see whether the rates will be significant.
	 * If the numbers did not agree then this test would be flawed, resulting in
	 * div by zero */
	if( x < 84. )
	{
		sexp_v = exp(-x);
	}
	else
	{
		sexp_v = 0.;
	}


#	ifdef DEBUG_FUN
	fputs( " <->sexp()\n", debug_fp );
#	endif
	return( sexp_v );
}

/*TestCode set flag saying that test code is in place 
 * prototype in cddefines.h */
void TestCode(void)
{

#	ifdef DEBUG_FUN
	fputs( "<+>TestCode()\n", debug_fp );
#	endif

	/* called if test code is in place */
	Testit.lgTestIt = TRUE;

#	ifdef DEBUG_FUN
	fputs( " <->TestCode()\n", debug_fp );
#	endif
	return;
}

/* sign.c - support for F77 sign() function translations by FOR_C (TM) */
/*
 *  copyright Cobalt Blue, Inc., 1988 - 1996
 *     ALL RIGHTS RESERVED
 
 * includes ssign (commented out(
 * isign (commented out)
 * 
 */
#define _ABS_(v)	( (v) < 0 ? -(v) : (v) )

#ifdef sign
#	undef sign
#endif

double sign(double a,double b )		/* floating point sign transfer */
{
	return( b < 0 ? -_ABS_(a) : _ABS_(a) );
}

/* min.c - support for F77 min() function variable arg transls by FOR_C (TM) */
/*
 *  copyright Cobalt Blue, Inc., 1988 - 1995
 *     ALL RIGHTS RESERVED
 */
#ifdef _MSC_VER
#	pragma warning( disable : 4244 )/* disable bogus conv int to short warning in MS VS*/
#	pragma warning( disable : 4056 )/* disable bogus underflow warning in MS VS*/
#endif

#define VA_START(a,v)	va_start(a,v)
#ifdef min
#	undef min
#endif

long min( long a, long b )
{
	return( a < b ? a : b );
}

#ifdef fmin
#	undef fmin
#endif

double fmin(double a, double b)
{
	return( a < b ? a : b );
}

/* NOTE:
		All of the functions in the rest of this file are,
		VARIABLE NUMBER OF ARGRUMENT FUNCTIONS!
 */


double vfmin( double mn, ... )
{
	double a;
	va_list args;	/* arg ptr */

	if( mn == FEND )
		return( 0. );	/* NULL arg list */

	VA_START(args,mn);	/* initialize variable args */

	while( (a=va_arg(args,double)) != FEND )
		mn = (mn < a) ? mn : a;

	return( mn );
}

/* max.c - support for F77 max() function variable arg transls by FOR_C (TM) */
/*
 *  copyright Cobalt Blue, Inc., 1988 - 1995
 *     ALL RIGHTS RESERVED
 */

/* NOTE: These functions guarantee max() problems won't occur due to
		side effects - at the cost of an extra function call. */
#ifdef smax
#	undef smax
#endif
short smax( short a, short b)
{
	return( a > b ? a : b );
}

#if 0
#ifdef max
#	undef max
#endif
long max( long a, long b )
{
	return( a > b ? a : b );
}
#endif

#ifdef fmax
#	undef fmax
#endif

double fmax(double a, double b)
{
	return( a > b ? a : b );
}

double vfmax( double mx, ... )
{
	double a;
	va_list args;	/* arg ptr */

	if( mx == FEND )
		return( 0. );	/* NULL arg list */

	VA_START(args,mx);	/* initialize variable args */

	while( (a=va_arg(args,double)) != FEND )
		mx = (mx > a) ? mx : a;

	return( mx );
}

/*broken set flag saying that the code is broken, */
void broken(void)
{

	broke.lgBroke = TRUE;

	return;
}

/*fixit - say that code needs to be fixed */
void fixit(void)
{

	broke.lgFixit = TRUE;

	return;
}

/* dprintf is a debug print routine that was provided by Peter Teuben,
 * as a component from his NEMO package.  It offers run-time specification
 * of the level of debugging */
void dprintf(int debug, const char *fmt, ...)
{
    va_list ap;

	/* print this debug message? (debug_level not currently used)*/
    if (debug <= trace.debug_level) 
	{		
		va_start(ap, fmt);	

		vfprintf(ioQQQ, fmt, ap);
		/* drain ioQQQ */
		fflush(ioQQQ);
		va_end(ap);
    }
}


/*qg32 32 point Gaussian quadrature, originally given to Gary F by Jim Lattimer */
double qg32(
	double xl, /*lower limit to integration range*/
	double xu, /*upper limit to integration range*/
	/*following is the pointer to the function that will be evaulated*/
	double (*fct)(double) )
{
	double a, 
	  b, 
	  c, 
	  y;


	/********************************************************************************
	 *                                                                              *
	 *  32-point Gaussian quadrature                                                *
	 *  xl  : the lower limit of integration                                        *
	 *  xu  : the upper limit                                                       *
	 *  fct : the (external) function                                               *
	 *  returns the value of the integral                                           *
	 *                                                                              *
	 * simple call to integrate sine from 0 to pi                                   *
	 * double agn = qg32( 0., 3.141592654 ,  sin );                                 *
	 *                                                                              *
	 *******************************************************************************/

	a = .5*(xu + xl);
	b = xu - xl;
	c = .498631930924740780*b;
	y = .35093050047350483e-2*((*fct)(a+c) + (*fct)(a-c));
	c = b*.49280575577263417;
	y += .8137197365452835e-2*((*fct)(a+c) + (*fct)(a-c));
	c = b*.48238112779375322;
	y += .1269603265463103e-1*((*fct)(a+c) + (*fct)(a-c));
	c = b*.46745303796886984;
	y += .17136931456510717e-1*((*fct)(a+c) + (*fct)(a-c));
	c = b*.44816057788302606;
	y += .21417949011113340e-1*((*fct)(a+c) + (*fct)(a-c));
	c = b*.42468380686628499;
	y += .25499029631188088e-1*((*fct)(a+c) + (*fct)(a-c));
	c = b*.3972418979839712;
	y += .29342046739267774e-1*((*fct)(a+c) + (*fct)(a-c));
	c = b*.36609105937014484;
	y += .32911111388180923e-1*((*fct)(a+c) + (*fct)(a-c));
	c = b*.3315221334651076;
	y += .36172897054424253e-1*((*fct)(a+c) + (*fct)(a-c));
	c = b*.29385787862038116;
	y += .39096947893535153e-1*((*fct)(a+c) + (*fct)(a-c));
	c = b*.2534499544661147;
	y += .41655962113473378e-1*((*fct)(a+c) + (*fct)(a-c));
	c = b*.21067563806531767;
	y += .43826046502201906e-1*((*fct)(a+c) + (*fct)(a-c));
	c = b*.16593430114106382;
	y += .45586939347881942e-1*((*fct)(a+c) + (*fct)(a-c));
	c = b*.11964368112606854;
	y += .46922199540402283e-1*((*fct)(a+c) + (*fct)(a-c));
	c = b*.7223598079139825e-1;
	y += .47819360039637430e-1*((*fct)(a+c) + (*fct)(a-c));
	c = b*.24153832843869158e-1;
	y = b*(y + .482700442573639e-1*((*fct)(a+c) + (*fct)(a-c)));

	/* the answer */

	return( y );
}

